/*********************************************************************
*
* (C) Copyright Broadcom Corporation 2000-2010
*
**********************************************************************
* @filename bootp_support.c
*
* @purpose dhcp main implementation
*
* @component bootp
*
* @comments none
*
* @create 6/26/2001
*
* @author paresh
*
* @end
*             
**********************************************************************/

/*************************************************************
                    
*************************************************************/




/*==========================================================================
  
        dhcp_prot.c
  
        DHCP protocol implementation for DHCP client
  
  ==========================================================================
  ####COPYRIGHTBEGIN####
                                                                            
   -------------------------------------------                              
   The contents of this file are subject to the Red Hat eCos Public License 
   Version 1.1 (the "License"); you may not use this file except in         
   compliance with the License.  You may obtain a copy of the License at    
   http:  www.redhat.com/                                                   
                                                                            
   Software distributed under the License is distributed on an "AS IS"      
   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the 
   License for the specific language governing rights and limitations under 
   the License.                                                             
                                                                            
   The Original Code is eCos - Embedded Configurable Operating System,      
   released September 30, 1998.                                             
                                                                            
   The Initial Developer of the Original Code is Red Hat.                   
   Portions created by Red Hat are                                          
   Copyright (C) 1998, 1999, 2000 Red Hat, Inc.                             
   All Rights Reserved.                                                     
   -------------------------------------------                              
                                                                            
  ####COPYRIGHTEND####
  ==========================================================================
  #####DESCRIPTIONBEGIN####
  
   Author(s):   hmt
   Contributors: gthomas
   Date:        2000-07-01
   Purpose:     DHCP support
   Description: 
  
  ####DESCRIPTIONEND####
  
  ========================================================================*/

#include "l7_common.h"
#include "datatypes.h"
#include "default_cnfgr.h"
#include "usmdb.h"
#include "osapi.h"
#include "osapi_support.h"
#include "simapi.h"
#include "log.h"
#include "sysapi.h"
#include "l7netapi.h"
#include "simapi.h"

#include "l7_socket.h"
/* this needs to get rolled into bsdsock.h */
#define ipcom_memset memset

#include <dhcp.h>

/* cant include dtlapi.h without getting socket.h */
#define L7_DTL_PORT_IF "dtl"

  #include <string.h>
  #include <unistd.h>
#include <sys/times.h>
#include <sys/types.h>
#include "bspapi.h"
#include "dns_client_api.h"

#ifdef _L7_OS_ECOS_
#include <sys/ioctl.h> 
#include <cyg/infra/diag.h>
#include <network.h>
#endif

L7_uint32 errNo;
L7_char8 *pErrString;

static L7_uchar8 timeout_random = 0;
extern struct in_addr previaddr;

struct timeout_state
{
  L7_uint32 secs;
  L7_int32 countdown;
};

static L7_uint32 dhcpcDebugMsgLevel = 0;

L7_uint32 dhcpcDebugMsgLvlSet(L7_uint32 level);
static void show_bootp(const L7_char8 *intf, struct bootp *bp);


/**************************************************************************
* @purpose  To get the actual packet size of an initialized buffer
*
* @param    bootp   DHCP structure
* 
*
* @returns  L7_int32 Size of the packet
*
* @comments None.
*
* @end
*************************************************************************/
static L7_int32 dhcp_size( struct bootp *ppkt )
{
  L7_uchar8 *op;
  L7_int32 size;

  op = &ppkt->bp_vend[0];       /* First check for the cookie!*/
  if ( op[0] !=  99 ||
       op[1] != 130 ||
       op[2] !=  83 ||
       op[3] !=  99 )
  {
    LOG_MSG( "dhcp_size():Bad DHCP cookie" );
    return 0;
  }
  op += 4;
  while ( *op != TAG_END )
  {
    op += *(op+1)+2;
    if ( op > &ppkt->bp_vend[BP_VEND_LEN-1] )
    {
      LOG_MSG( "dhcp_size(): Oversize DHCP packet in dhcp_size" );
      return 0;
    }
  }
  size=  op - (L7_uchar8 *)ppkt + 1;
  return size;
}

/**************************************************************************
* @purpose  To remove a bootp option value if it is present from an 
*           initialized buffer
*
* @param     bootp      DHCP structure
* @param     L7_uchar8    Option identifier
* @param     L7_uint32      Option Value
* @param     L7_int32       Option Value Length
*
* @returns  L7_SUCCESS or L7_FAILURE
*
* @comments None.
*
* @end
*************************************************************************/
static L7_int32
remove_tag( struct bootp *ppkt,
               L7_uchar8 tag)
{
  L7_uchar8 *op, *begin, *end;
  L7_uint32 len;

  /* Initially this will only scan the options field.*/

  op = &ppkt->bp_vend[4];

  /*searching for option by looking at all the option tag values*/
  while ( *op != TAG_END )
  {
    if ( op > &ppkt->bp_vend[BP_VEND_LEN-1] )
    {
      LOG_MSG( "set_fixed_tag():Oversize DHCP packet in set_fixed_tag" );
      return L7_FAILURE; 
    }
    if ( *op == tag )
      break;
    op += *(op+1)+2;
  }

  /*option was found, so remove it*/
  if ( *op == tag )
  {
    begin = op;
    /*move pointer past the tag to be removed*/
    op += *(op+1)+2;
    end = op; /*begin and end indicate the start and end of option to be removed*/

    while ( *op != TAG_END )
    {
        if ( op > &ppkt->bp_vend[BP_VEND_LEN-1] )
        {
          LOG_MSG( "set_fixed_tag():Oversize DHCP packet in set_fixed_tag" );
          return L7_FAILURE; 
        }
        op += *(op+1)+2;
    }
    len = op - end + 1;        /*extra 1 to take into account the TAG_END byte*/
    memmove(begin, end, len); 
    return L7_SUCCESS;
  }

  /*option was not present in the first place*/
  return L7_SUCCESS;
}

/* ------------------------------------------------------------------------
   Insert/set an option value in an initialized buffer*/


/**************************************************************************
* @purpose  To insert/set a bootp option valure in initialized buffer
*
* @param     bootp      DHCP structure
* @param     L7_uchar8    Option identifier
* @param     L7_uint32      Option Value
* @param     L7_int32       Option Valure Length
*
* @returns  L7_TRUE or L7_FALSE
*
* @comments None.
*
* @end
*************************************************************************/
static L7_int32
set_fixed_tag( struct bootp *ppkt,
               L7_uchar8 tag,
               L7_uint32 value,
               L7_int32 len)
{
  L7_uchar8 *op;

  /* Initially this will only scan the options field.*/

  op = &ppkt->bp_vend[4];
  while ( *op != TAG_END )
  {
    if ( op > &ppkt->bp_vend[BP_VEND_LEN-1] )
    {
      LOG_MSG( "set_fixed_tag():Oversize DHCP packet in set_fixed_tag" );
      return L7_FALSE; 
    }
    if ( *op == tag )
      break;
    op += *(op+1)+2;
  }

  if ( *op == tag )
  {
    if ( *(op+1) != len )
    {
      LOG_MSG( "set_fixed_tag():Wrong size in set_fixed_tag" );
      return L7_FALSE;          
    }
  }
  else
  { /* overwrite the end tag and install a new one*/
    if ( op + len + 2 > &ppkt->bp_vend[BP_VEND_LEN-1] )
    {
      LOG_MSG( "set_fixed_tag():Oversize DHCP packet in set_fixed_tag append" );
      return L7_FALSE; 
    }
    *op = tag;
    *(op+1) = len;
    *(op + len + 2) = TAG_END;
  }
  /* and insert the value.  Net order is BE.*/
  op += len + 2 - 1;              /* point to end of value*/
  while ( len-- > 0 )
  {
    *op-- = (L7_uchar8)(value & 255);
    value >>= 8;
  }
  return L7_TRUE;
}

/**************************************************************************
* @purpose  To add a bootp option value in initialized buffer
*
* @param     bootp        DHCP structure
* @param     L7_uchar8    Option identifier
* @param     L7_uchar8    Option Value
*
* @returns  L7_TRUE or L7_FALSE
*
* @comments None.
*
* @end
*************************************************************************/
static L7_int32
add_fixed_tag( struct bootp *ppkt,
               L7_uchar8 addingTag,
               L7_uchar8 value)
{
  struct bp_option *tagEndPtr = L7_NULLPTR;
  struct bp_option *bp_TagPtr = L7_NULLPTR;

  /* Initially this will only scan the options field.*/
  tagEndPtr = bp_TagPtr = (struct bp_option*)&ppkt->bp_vend[4];

  /* first find the end of options list */
  while( tagEndPtr->tag != TAG_END )
  {
    /* check whether we are out of the options boundary */
    if ( (L7_uchar8*)tagEndPtr > &ppkt->bp_vend[BP_VEND_LEN-1] )
    {
      LOG_MSG( "add_fixed_tag():Oversize DHCP packet in add_fixed_tag (%d)", addingTag );
      return L7_FALSE;
    }
    tagEndPtr = (struct bp_option *)(&(tagEndPtr->data) + tagEndPtr->dataLen);  /* jump to next existing TAG */
  }

  /* find option to set. Continue while TAG_END isn't met */
  while ( bp_TagPtr->tag != TAG_END )  
  {
    /* check whether we are out of the options boundary */
    if ( (L7_uchar8*)bp_TagPtr > &ppkt->bp_vend[BP_VEND_LEN-1] )
    {
      LOG_MSG( "add_fixed_tag():Oversize DHCP packet in add_fixed_tag(%d)", addingTag );
      return L7_FALSE;
    }
    if ( bp_TagPtr->tag == addingTag )
      break;
    bp_TagPtr = (struct bp_option *)(&(bp_TagPtr->data) + bp_TagPtr->dataLen);  /* jump to next existing TAG */
  }

  /* if option already exists */
  if ( bp_TagPtr->tag == addingTag )
  {
    if( bp_TagPtr->tag == TAG_DHCP_PARM_REQ_LIST )
    {
      L7_uint32 optCounter;
      
      /* if option value to add already exists in option fields do return */
      for(optCounter = 0; optCounter < bp_TagPtr->dataLen; optCounter++)
      {
        if( value == *( &(bp_TagPtr->data) + optCounter) )
        {
          return L7_TRUE;
        }
      }
      /* otherwise install new value in option fields. 
         Check whether adding new option value would cause oversize of packet */
      if( (L7_uchar8*)tagEndPtr + 1 > &ppkt->bp_vend[ BP_VEND_LEN-1] )
      {
        LOG_MSG( "add_fixed_tag():Oversize DHCP packet in add_fixed_tag(%d)", addingTag );
        return L7_FALSE;
      }
      /* make empty field for insertion by moving existing data fields and TAG_END field */
      memmove( &(bp_TagPtr->data) + 1, &(bp_TagPtr->data), (L7_uchar8*)tagEndPtr - (L7_uchar8*)bp_TagPtr - 1);
      /* increase length field */
      bp_TagPtr->dataLen += 1;
    }
  }
  /* if option doesn't already exist - insert a new option before the TAG_END */
  else
  {
    /* check if adding new option tag, length, value fields wouldn't cause oversize of packet */
    if( (L7_uchar8*)tagEndPtr + 3 > &ppkt->bp_vend[ BP_VEND_LEN-1] )
    {
      LOG_MSG( "add_fixed_tag():Oversize DHCP packet in add_fixed_tag(%d)", addingTag );
      return L7_FALSE;
    }
    bp_TagPtr->tag = addingTag;        /* set new tag */
    bp_TagPtr->dataLen = 1;            /* set data field length */
    *( &(bp_TagPtr->data) + 1 ) = TAG_END; /* overwrite the end tag */
  }
  /* and insert the value.  Net order is BE.*/
  bp_TagPtr->data = value;

  return L7_TRUE;
}



/**************************************************************************
* @purpose  Initialize the timer values
*
* @param     timeval        Ptr to timeval structure
* @param     timeout_state  Ptr to state variable
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
static void reset_timeout( struct timeval *ptv, struct timeout_state *pstate )
{
  timeout_random++;
  pstate->countdown = 4; /* initial fast retries*/
  pstate->secs = 4 + (timeout_random & 3);
  ptv->tv_sec = 4;
  ptv->tv_usec = 65536 * (2 + (timeout_random & 3));
}


/**************************************************************************
* @purpose   To  find out next timer value
*
* @param     timeval        Ptr to timeval structure
* @param     timeout_state  Ptr to state variable
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
static L7_int32 next_timeout( struct timeval *ptv, struct timeout_state *pstate )
{
  if ( 0 > pstate->countdown-- )
    return L7_TRUE;
  timeout_random++;
  pstate->secs = (ptv->tv_sec * 2) - 2 + (timeout_random & 3);
  pstate->countdown = 2; 
  ptv->tv_sec = pstate->secs ;
  return pstate->secs < 50; /* If longer, too many tries...*/
}


/**************************************************************************
* @purpose  Alarm to notify T1, T2 timers and lease expiry
*
* @param    L7_uint32      Address for dhcp_lease structure
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
static void alarm_function(L7_uint32 data)
{
  struct dhcp_lease *lease = (struct dhcp_lease *)data;

  lease->alarm = L7_NULLPTR;
  lease->which |= lease->next;

  /* Step the lease on into its next state of being alarmed ;-)*/
  if ( lease->next & DHCP_LEASE_EX )
  {
    *(lease->dhcpState)=DHCPSTATE_DO_RELEASE;
    *(lease->configStatus) = L7_SYSCONFIG_MODE_NOT_COMPLETE;    
  }
  else if ( lease->next & DHCP_LEASE_T2 )
  {
    osapiTimerAdd((void*) alarm_function, (L7_uint32)lease, L7_NULL, lease->expiry , &lease->alarm );
    lease->next = DHCP_LEASE_EX;
    *(lease->dhcpState)= DHCPSTATE_REBINDING;
    *(lease->configStatus) = L7_SYSCONFIG_MODE_NOT_COMPLETE;    
  }
  else if ( lease->next & DHCP_LEASE_T1 )
  {
    osapiTimerAdd((void*) alarm_function, (L7_uint32)lease, L7_NULL, lease->t2 , &lease->alarm );
    lease->next = DHCP_LEASE_T2;
    *(lease->dhcpState)= DHCPSTATE_RENEWING;
    *(lease->configStatus) = L7_SYSCONFIG_MODE_NOT_COMPLETE;                                    
  }
}


/**************************************************************************
* @purpose  Read lease renew / rebind / expiry times from packet
*
* @param    bootp*             Ptr to dhcp packet
* @param    dhcp_lease*        Ptr to lease structure
*
* @returns  Void
*
* @comments Added for defect 4356.
*
* @end
*************************************************************************/
static void get_lease_params( struct bootp *bootp, struct dhcp_lease *lease )
{
  L7_uint32 pkt_lease_time = 0;
  L7_uint32 pkt_t1 = 0;
  L7_uint32 pkt_t2 = 0;

  /* 
   * Lease time MUST be included (for DHCPOFFER and DHCPACK in response to 
   * a DHCPREQUEST) so it's okay to assume it's there.
   */
  l7_get_bootp_option( bootp, TAG_DHCP_LEASE_TIME, &pkt_lease_time, sizeof(pkt_lease_time) );
  /* Retrieve T1 and T2 from packet, if specified */
  l7_get_bootp_option( bootp, TAG_DHCP_RENEWAL_TIME, &pkt_t1, sizeof(pkt_t1) );
  l7_get_bootp_option( bootp, TAG_DHCP_REBIND_TIME, &pkt_t2, sizeof(pkt_t2) );
  
  /* If not specified, values will be 0 */
  if ( 0 != pkt_lease_time) {
    lease->net_expiry = pkt_lease_time;
  }
  if ( 0 != pkt_t1 ) {
    lease->net_t1 = pkt_t1;
  }
  if ( 0 != pkt_t2 ) {
    lease->net_t2 = pkt_t2;
  }
  if ((osapiNtohl(lease->net_t1) >= osapiNtohl(lease->net_t2)) || 
      (osapiNtohl(lease->net_t1) >= osapiNtohl(lease->expiry))) {
    lease->net_t1 = 0;
  }
  if (osapiNtohl(lease->net_t2) >= osapiNtohl(lease->expiry)) {
    lease->net_t2 = 0;
  }
}


/**************************************************************************
* @purpose  Delete timer from any existing lease
*
* @param    dhcp_lease*        Ptr to lease structure
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
static void del_lease_timer( struct dhcp_lease *lease )
{
  osapiTimerDescr_t *pTimerHolder;
  pTimerHolder= lease->alarm;
  if ( pTimerHolder )
  {
    osapiTimerFree(pTimerHolder);  /* Destroy the timer */
    pTimerHolder = 0;
    lease->alarm = L7_NULLPTR;
  }
}


/**************************************************************************
* @purpose  Initialize the lease structure
*
* @param    dhcp_lease*        Ptr to lease structure
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
static void no_lease( struct dhcp_lease *lease )
{
  del_lease_timer( lease );
  lease->net_t1 = lease->net_t2 = lease->net_expiry = 0;
}


/**************************************************************************
* @purpose  Initialize the new lease structure values
*
* @param     dhcp_lease*    Ptr to lease structure
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
static void new_lease( struct dhcp_lease *lease )
{
  L7_uint32 now ;/*= cyg_current_time();*/
  L7_uint32 then;

  osapiTimerDescr_t *pTimerHolder;

  /*now=osapiUpTimeRaw();*/
  now=0;
  pTimerHolder= lease->alarm;

  /* Silence any jabbering from past lease on this interface*/
  del_lease_timer( lease );
  lease->which = lease->next = 0;

  if ( 0xffffffff == lease->net_expiry )
  {
    lease->expiry = 0xffffffff;
    lease->t2     = 0xffffffff;
    lease->t1     = 0xffffffff;
    return; /* it's an infinite lease*/
  }
  
  /* Calculate T1 and T2's default values if they weren't specified */
  if ( 0 == lease->net_t1 ) {
      lease->net_t1 = osapiHtonl(osapiNtohl(lease->net_expiry) / 2);
  }
  if ( 0 == lease->net_t2 ) {
      lease->net_t2 = osapiHtonl((osapiNtohl(lease->net_expiry) / 8) * 7);
  }

  /*set T1 timer value*/
  then = (L7_double64)(osapiNtohl(lease->net_t1));
  then *= 1000; /* into mS*/
  lease->t1 = now + then;

  /*set T2 timer value*/
  then = (L7_double64)(osapiNtohl(lease->net_t2));
  then *= 1000; /* into mS - we know there is room in a tick_count_t*/
  lease->t2 = now + then - lease->t1; /*relative time */

  /* extract the lease time and scale it &c to now.*/
  then = (L7_long32)(osapiNtohl(lease->net_expiry));
  then *= 1000; 
  lease->expiry = now + then - (lease->t2 + lease->t1); /*relative time*/

#if 0 /* for testing this mechanism*/
  lease->expiry = now + 5000; /* 1000 here makes for failure in the DHCP test*/
  lease->t2     = now + 3500;
  lease->t1     = now + 2500;
#endif

  lease->next = DHCP_LEASE_T1;
  osapiTimerAdd((void*) alarm_function, (L7_uint32) lease, L7_NULL, lease->t1 , &lease->alarm);    
}


/**************************************************************************
* @purpose  Add DNS options to DNS Client
*
* @param    received @b{(input)} Ptr to boot structure
*
* @returns  void
*
* @comments none
*
* @end
*************************************************************************/
static void dhcpClientDnsOptionsAdd(struct bootp *received)
{
  L7_uint32  serverIp[L7_DNS_NAME_SERVER_ENTRIES];
  L7_uchar8  domainNameList[L7_DNS_DOMAIN_LIST_NAME_SIZE_MAX + 1];

  memset(domainNameList, 0, sizeof(domainNameList));
  memset(serverIp, 0, sizeof(serverIp));

  if (l7_get_bootp_option (received, TAG_DOMAIN_SERVER, &serverIp, sizeof(serverIp)) == L7_TRUE)
  {
    /* if domain name sever option is found in response, add it to dns name server list. */
    L7_uint32 i=0;
    for (i=0; i<L7_DNS_NAME_SERVER_ENTRIES;i++)
    {
      if (serverIp[i] == 0)
      {
        break;
      }

      if (dnsClientNameServerEntryAdd(serverIp[i]) != L7_SUCCESS)
      {
        LOG_MSG("dhcpClientDnsOptionsAdd:  DNS name server entry add failed.");
      }
    }
  }

  if (l7_get_bootp_option (received, TAG_DOMAIN_NAME, domainNameList, sizeof(domainNameList)) == L7_TRUE)
  {
    /* if domain name List option is found in response, add it to dns domain name List. */
    if (dnsClientDomainNameListAdd(domainNameList) != L7_SUCCESS)
    {
      LOG_MSG("dhcpClientDnsOptionsAdd: DNS domain name list entry add failed.");
    }
  }
}

/**************************************************************************
* @purpose  Remove DNS options from DNS Client
*
* @param    xmit @b{(input)} Ptr to boot structure
*
* @returns  void
*
* @comments none
*
* @end
*************************************************************************/
static void dhcpClientDnsOptionsRemove(struct bootp *xmit)
{
  L7_uint32  serverIp[L7_DNS_NAME_SERVER_ENTRIES];
  L7_uchar8  domainNameList[L7_DNS_DOMAIN_LIST_NAME_SIZE_MAX + 1];

  memset(domainNameList, 0, sizeof(domainNameList));
  memset(serverIp, 0, sizeof(serverIp));

  if (l7_get_bootp_option (xmit, TAG_DOMAIN_SERVER, &serverIp, sizeof(serverIp)) == L7_TRUE)
  {
    /* if domain name sever option is found in response,
     * delete from dns name server list. 
     */
    L7_uint32 i=0;
    for (i=0; i<L7_DNS_NAME_SERVER_ENTRIES;i++)
    {
      if (serverIp[i] == 0)
      {
        break;
      }

      if (dnsClientNameServerEntryRemove(serverIp[i]) != L7_SUCCESS)
      {
        LOG_MSG("dhcpClientDnsOptionsRemove:  DNS name server entry delete failed.");
      }
    }
  }

  if (l7_get_bootp_option (xmit, TAG_DOMAIN_NAME, domainNameList, sizeof(domainNameList)) == L7_TRUE)
  {
    /* if domain name List option is found in response,
     * delete from dns domain name List. 
     */
    if (dnsClientDomainNameListRemove(domainNameList) != L7_SUCCESS)
    {
      LOG_MSG("dhcpClientDnsOptionsRemove: DNS domain name list entry delete failed.");
    }
  }
}

/**************************************************************************
* @purpose   the DHCP state machine - this does all the work
*
* @param     L7_uchar8*     Interface name
* @param     struct bootp*    Ptr to boot structure
* @param     L7_short16*    Ptr to DHCP state
* @param     struct dhcp_lease* Ptr to lease structure
* @param   L7_short16*    Ptr to configuration state
*
* @returns  L7_TRUE or L7_FALSE
*
* @comments None.
*
* @end
*************************************************************************/
L7_int32
do_dhcp(const L7_uchar8 *intf, struct bootp *res,
        L7_short16 *pstate, struct dhcp_lease *lease,L7_short16* pConfigState)
{
  L7_sockaddr_in_t cli_addr, broadcast_addr, server_addr, rx_addr;
  L7_int32 s, addrlen, maxFdLength;
  L7_int32 one = 1;
  L7_char8 TEN = 10;
  L7_uchar8 mincookie[] = {99,130,83,99,255} ;
  struct timeval tv;
  struct timeout_state timeout_scratch;
  L7_ushort16 oldstate = *pstate;
  L7_uchar8 msgtype = 0, seen_bootp_reply = 0;
  L7_uchar8 * macAddress[BP_CHADDR_LEN];
  struct ifreq ifr;
  L7_uint32 xid=L7_NULL;
  fd_set  pReadFd;
  L7_short16 selectCounter=0, requestCounter=0, renewCounter=0, rebindCounter=0;
  L7_uchar8 ifname[IFNAMSIZ];
  L7_uchar8 sysBIA[L7_MAC_ADDR_LEN];     /* for xid randomness */
  L7_uchar8 srvportBIA[L7_MAC_ADDR_LEN]; /* for xid randomness */
  L7_uint32       servPortConfigMode, networkConfigMode, currentIpAddr;
  L7_uint32 errorMsg = L7_NULL, ifNetMask, logMsg = L7_NULL;
  L7_uint32 bsent,brecvd;
  L7_char8   addrStr[OSAPI_INET_NTOA_BUF_SIZE];
  struct in_addr   local_bp_yiaddr;     /* client IP address */
  struct in_addr   local_bp_siaddr;     /* server IP address */
  struct bootp_serverinfo rx_local;

#ifdef _L7_OS_ECOS_
  struct sockaddr_in *addrp;
  struct ecos_rtentry route;
#endif

  /* IMPORTANT: xmit is the same as res throughout this; *received is a
     scratch buffer for reception; its contents are always copied to res
     when we are happy with them.  So we always transmit from the
     existing state.*/
   struct bootp *received = (struct bootp *)&rx_local;

  struct bootp *xmit = res;
  *pConfigState= CONFIG_NONE;

  memset(ifname, 0, IFNAMSIZ);
  memset(&rx_local, 0, sizeof(struct bootp_serverinfo));

  /*memset(ifname, 0, 20);*/
  if ( strcmp(bspapiServicePortNameGet(), intf)==0 )
      sprintf(ifname, "%s0", bspapiServicePortNameGet());
  else if ( strcmp(L7_DTL_PORT_IF, intf)==0 )
      sprintf(ifname, "%s0", L7_DTL_PORT_IF);

  /* First, get a socket on the interface in question.  But Zeroth, if
     needs be, bring it to the half-up broadcast only state if needs be.*/

  if ( DHCPSTATE_INIT        == oldstate
       || DHCPSTATE_FAILED   == oldstate
       || DHCPSTATE_NOTBOUND == oldstate
       || 0                  == oldstate )
  {
    /*
    DHCPServicePortIfSet();
    */

    /*
    printf("DHCP intf %s, Service Port %s\n", intf, bspapiServicePortNameGet());
    */

	dhcpConfigSemGet ();
    if ( strcmp(bspapiServicePortNameGet(), intf)==0 )
    {
      /*
      printf("DHCP using Service port %s0\n", bspapiServicePortNameGet());
      */
      simSetServPortIPAddr(0,L7_FALSE);
      osapiNetIPSet(ifname, 0);

      /* If we ever run DHCP on both interfaces at the same time, then distiguish. */
      simGetServicePortBurnedInMac(&srvportBIA[0]);
      xid = (L7_uint32)res + 
            (L7_uint32) (0xffffffff & osapiUpTimeRaw()) + 
            (L7_uint32)(srvportBIA[2]<<24) + 
            (L7_uint32)(srvportBIA[3]<<16) + 
            (L7_uint32)(srvportBIA[4]<<8) + 
            (L7_uint32)(srvportBIA[5]); 
    }
    else if ( strcmp(L7_DTL_PORT_IF, intf)==0 )
    {
      /*
      */
      simSetSystemIPAddr(FD_SIM_DEFAULT_NETWORK_IP_ADDR_NETGEAR, L7_FALSE);
      simSetSystemIPNetMask(FD_SIM_DEFAULT_NETWORK_SUBNET_MASK_NETGEAR);
     simSetSystemIPGateway(FD_SIM_DEFAULT_NETWORK_GATEWAY_ADDR_NETGEAR, L7_FALSE);

#ifndef VXWORKS
      fflush(0);
#endif

      /* If we ever run DHCP on both interfaces at the same time, then distiguish. */
      simGetSystemIPBurnedInMac(&sysBIA[0]);
      xid = (L7_uint32)res + 
            (L7_uint32) (0xffffffff & osapiUpTimeRaw()) + 
            (L7_uint32)(sysBIA[2]<<24) + 
            (L7_uint32)(sysBIA[3]<<16) + 
            (L7_uint32)(sysBIA[4]<<8) + 
            (L7_uint32)(sysBIA[5]); 
    }
	dhcpConfigSemFree ();

    sleep(5);

#ifdef _L7_OS_LINUX_
    /* This is required on linux, but harmful on VxWorks. VxWorks installs a hidden
     * route to 255.255.255.255 at startup. Adding the same route here would either fail
     * or just bump the ref count in the stack's routing table. Then if DHCP fails 
     * for example if the mgmt interface is down, removing this route will remove
     * the route installed by VxWorks itself. Then sending to 255.255.255.255 will 
     * never work again. One might wonder why you would even do a routing table lookup
     * for a broadcast packet, but there it is. */
    osapimRouteEntryAdd (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif

    *pstate = DHCPSTATE_INIT;

  }
  else
    xid = osapiNtohl(res->bp_xid); /*use what's there already*/

  if (osapiSocketCreate(L7_AF_INET, L7_SOCK_DGRAM, 0, &s) == L7_FAILURE)
  {
    LOG_MSG("do_dhcp(): Failed to create socket");
    /* Delete route entry for 0.0.0.0 IP address */
#ifdef _L7_OS_LINUX_
    osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
    return L7_FALSE;
  }

  FD_ZERO(&pReadFd);
  FD_SET(s, &pReadFd);
  maxFdLength= s +1;

  if ( osapiSetsockopt(s, L7_SOL_SOCKET, L7_SO_REUSEADDR,(L7_char8 *) &one, sizeof(one)) == L7_FAILURE )
  {
    LOG_MSG("do_dhcp(): osapiSetsockopt SO_REUSEADDR failed");
    /*delete route entry for 0.0.0.0 IP address */
#ifdef _L7_OS_LINUX_
    osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
    osapiSocketClose (s);
    return L7_FALSE;
  }

#ifdef _L7_OS_VXWORKS_
  if ( osapiSetsockopt(s, L7_SOL_SOCKET, L7_SO_BINDTODEVICE, ifname, IFNAMSIZ) )
  {
    LOG_MSG("do_dhcp(): osapiSetsockopt SO_BINDTODEVICE failed");
    /*delete route entry for 0.0.0.0 IP address */
    osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
    osapiSocketClose (s);
    return L7_FALSE;
  }
#endif


  if ( DHCPSTATE_INIT        == oldstate
       || DHCPSTATE_FAILED   == oldstate
       || DHCPSTATE_NOTBOUND == oldstate
       || 0                  == oldstate
       || DHCPSTATE_REBINDING == oldstate )
  {
    sprintf(ifr.ifr_name,"%s%s", intf,"0");/* Add device number, zero only for 
                          first interface device of X type  */
    bzero ((caddr_t) &ifr.ifr_addr, sizeof (ifr.ifr_addr));  

    ifr.ifr_addr.sa_family = AF_INET;  

    if (osapiSetsockopt(s, L7_SOL_SOCKET, L7_SO_BROADCAST,(L7_char8 *) &one, sizeof(one))== L7_FAILURE)
    {
      LOG_MSG("do_dhcp(): osapiSetsockopt SO_BROADCAST failed");
#ifdef _L7_OS_LINUX_
      /*delete route entry for 0.0.0.0 IP address */
      osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
      osapiSocketClose (s);
      return L7_FALSE;
    }

#ifdef _L7_OS_ECOS_
    /*cyg_route_reinit();*/
    addrp = (struct sockaddr_in *) &ifr.ifr_addr;
    memset(addrp, 0, sizeof(*addrp));
    addrp->sin_family = AF_INET;
    addrp->sin_len = sizeof(*addrp);
    addrp->sin_port = 0;
    addrp->sin_addr.s_addr = FD_SIM_DEFAULT_NETWORK_IP_ADDR_NETGEAR;

    if (ioctl(s, SIOCSIFADDR, &ifr)) {
        perror("SIOCSIFADDR");
        return false;
    }
    if (ioctl(s, SIOCSIFNETMASK, &ifr)) {
        perror("SIOCSIFNETMASK");
        return false;
    }
    if (ioctl(s, SIOCSIFADDR, &ifr)) {
        perror("SIOCSIFADDR");
        return false;
    }

    /* the broadcast address is 255.255.255.255 */
    memset(&addrp->sin_addr, 255, sizeof(addrp->sin_addr));
    if (ioctl(s, SIOCSIFBRDADDR, &ifr)) {
        perror("SIOCSIFBRDADDR");
        return false;
    }
    // Set up routing
    /* the broadcast address is 255.255.255.255 */
    memset(&addrp->sin_addr, 255, sizeof(addrp->sin_addr));
    memset(&route, 0, sizeof(route));
    memcpy(&route.rt_gateway, addrp, sizeof(*addrp));

    addrp->sin_family = AF_INET;
    addrp->sin_port = 0;
    addrp->sin_addr.s_addr = INADDR_ANY;
    memcpy(&route.rt_dst, addrp, sizeof(*addrp));
    memcpy(&route.rt_genmask, addrp, sizeof(*addrp));

    route.rt_dev = ifr.ifr_name;
    route.rt_flags = RTF_UP|RTF_GATEWAY;
    route.rt_metric = 0;

    if (ioctl(s, SIOCADDRT, &route)) {
        if (errno != EEXIST) {
            perror("SIOCADDRT 3");
            return false;
        }
    }

#endif
    ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr = osapiHtonl(0xffffffff);

      /* Set the mcast TTL. Relay agents, like FASTPATH's own relay agent,
       * may retain the TTL in the unicast packets they relay. mcast TTL 
       * defaults to 1, meaning relay agent can't relay. */
      if ( osapiSetsockopt(s, IPPROTO_IP, L7_IP_MULTICAST_TTL,
                           (L7_char8 *) &TEN, sizeof(TEN)) == L7_FAILURE )
      {
        LOG_MSG("do_dhcp(): osapiSetsockopt IP_MULTICAST_TTL failed");
#ifdef _L7_OS_LINUX_
        /*delete route entry for 0.0.0.0 IP address */
        osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
        osapiSocketClose (s);
        return L7_FALSE;
      }
  }

  memset((L7_char8 *) &cli_addr, 0, sizeof(cli_addr));
  cli_addr.sin_family = L7_AF_INET;
  cli_addr.sin_addr.s_addr = osapiHtonl(INADDR_ANY);
  cli_addr.sin_port = osapiHtons(IPPORT_BOOTPC);

  memset((L7_char8 *) &broadcast_addr, 0, sizeof(broadcast_addr));
  broadcast_addr.sin_family = L7_AF_INET;
  broadcast_addr.sin_addr.s_addr = osapiHtonl(0xffffffff);
  broadcast_addr.sin_port = osapiHtons(IPPORT_BOOTPS);

  memset((L7_char8 *) &server_addr, 0, sizeof(server_addr));
  server_addr.sin_family = L7_AF_INET;
  server_addr.sin_addr.s_addr = osapiHtonl(INADDR_BROADCAST); /* overwrite later*/
  server_addr.sin_port = osapiHtons(IPPORT_BOOTPS);

  if(osapiSocketBind(s, (L7_sockaddr_t *)&cli_addr, sizeof(cli_addr)) < 0)
  {
    LOG_MSG("do_dhcp(): osapiSocketBind failed");
#ifdef _L7_OS_LINUX_
    /*delete route entry for 0.0.0.0 IP address */
    osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
    osapiSocketClose (s);
    return L7_FALSE;
  }

  /* Now, we can launch into the DHCP state machine.  I think this will
     be the neatest way to do it; it returns from within the switch arms
     when all is well, or utterly failed.*/

  reset_timeout( &tv, &timeout_scratch );

  while ( 1 )
  {
    switch ( *pstate )
    {
      
      case DHCPSTATE_INIT:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP INIT state\n");
        }

        /* need to reset when dhcp fails and retries */
        FD_ZERO(&pReadFd);
        FD_SET(s, &pReadFd);

        /* Send the DHCPDISCOVER packet*/

        dhcpConfigSemGet ();
        if ( strcmp(bspapiServicePortNameGet(), intf)==0 )
        {
          simGetServicePortBurnedInMac((L7_uchar8 *) macAddress);
        }
        else if ( strcmp(L7_DTL_PORT_IF, intf)==0 )
        {
          if ( simGetSystemIPMacType() == L7_SYSMAC_BIA )
            simGetSystemIPBurnedInMac((L7_uchar8 *) macAddress);
          else
            simGetSystemIPLocalAdminMac((L7_uchar8 *) macAddress);
        }
        dhcpConfigSemFree ();

        /* Fill in the BOOTP request - DHCPDISCOVER packet*/
        bzero((char *)xmit, sizeof(*xmit));
        xmit->bp_op = BOOTREQUEST;
        xmit->bp_htype = HTYPE_ETHERNET;
        xmit->bp_hlen = L7_MAC_ADDR_LEN;
        xmit->bp_xid = osapiHtonl(xid);
        xmit->bp_secs = 0;
        xmit->bp_flags = osapiHtons(0x8000); /* BROADCAST FLAG*/

        bcopy((L7_uchar8 *) macAddress, (L7_uchar8 *)&xmit->bp_chaddr, xmit->bp_hlen);
        bcopy(mincookie, xmit->bp_vend, sizeof(mincookie));

        /* Set DHCP DISCOVER int the packet*/
        set_fixed_tag( xmit, TAG_DHCP_MESS_TYPE, DHCPDISCOVER, 1 );
        if ( previaddr.s_addr != 0 )
          /* Request the previous IP address, would like to have the same*/
          set_fixed_tag( xmit, TAG_DHCP_REQ_IP, osapiNtohl(previaddr.s_addr), 4 );

        set_fixed_tag( xmit, TAG_DHCP_MAX_MSGSZ, BP_MINPKTSZ, 2 );
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_DOMAIN_SERVER);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_GATEWAY);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_SUBNET_MASK);
  
        if (dhcpcDebugMsgLevel > 0)
        {
          printf( "---------DHCPSTATE_INIT sending:\n" );
          show_bootp( intf, xmit );
        }   

        if ( osapiSocketSendto(s, (L7_char8 *)xmit, dhcp_size(xmit), 0, 
                    (L7_sockaddr_t *)&broadcast_addr, sizeof(broadcast_addr),&bsent)  != L7_SUCCESS )
        {

          /*
          printf("Could not send to 0x%x\n",
		         osapiNtohl(broadcast_addr.sin_addr.s_addr));
          */
          errNo=osapiErrnoGet();
          pErrString=strerror(errNo);
          printf("\nosapiSocketSendto() failed in do_dhcp() with error %d, %s.", errNo, pErrString);
          *pstate = DHCPSTATE_FAILED;
          break;
        }

        seen_bootp_reply = 0;
        *pstate = DHCPSTATE_SELECTING;
        break;

      case DHCPSTATE_SELECTING:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP SELECTING state\n");
        }

        /* This is a separate state so that we can listen again
           without* retransmitting.listen for the DHCPOFFER reply*/

        addrlen = sizeof(rx_addr);

        if (dhcpcDebugMsgLevel > 0)
        {
          printf("DHCP SELECTING: calling select, timeout=%ld\n", tv.tv_sec);
        }

        if ( osapiSelect(maxFdLength, &pReadFd, NULL, NULL, tv.tv_sec, tv.tv_usec) <= 0 )
        {
          if ( ! next_timeout( &tv, &timeout_scratch ) )
          {
            *pstate = DHCPSTATE_FAILED;
            break;
          }
          *pstate = DHCPSTATE_INIT; /* to retransmit*/
          break;
        }

        if ( osapiSocketRecvfrom(s,(L7_char8*) received, sizeof(struct bootp), MSG_DONTWAIT,
                      ( L7_sockaddr_t *)&rx_addr, &addrlen, &brecvd)  != L7_SUCCESS )
        {
          /* No packet arrived (this time)*/
          if ( seen_bootp_reply )
          { /* then already have a bootp reply
            / Save the good packet in *xmit*/
            bcopy((L7_char8*) received ,(L7_char8*) xmit, dhcp_size(received) );
            *pstate = DHCPSTATE_BOOTP_FALLBACK;
            reset_timeout( &tv, &timeout_scratch );
            break;
          }
          /* go to the next larger timeout and re-send:*/
          if ( ! next_timeout( &tv, &timeout_scratch ) )
          {
            *pstate = DHCPSTATE_FAILED;
            break;
          }
          *pstate = DHCPSTATE_INIT; /* to retransmit*/
          break;
        }

        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(rx_addr.sin_addr.s_addr), addrStr);
          printf( "---------DHCPSTATE_SELECTING received:\n" );
          printf( "...rx_addr is family %d, addr %s, port %d\n",
                  rx_addr.sin_family,
                  addrStr,
                  osapiNtohs(rx_addr.sin_port) );
          show_bootp( intf, received );
        }
            
        if ( received->bp_xid != osapiHtonl(xid) )
        {   /* not the same transaction;*/
          if ( selectCounter < 4 )
            selectCounter++;
          else
            *pstate = DHCPSTATE_FAILED;

          break;      /* listen again...    */
        }

        if ( 0 == received->bp_siaddr.s_addr )
        {
          /* then fill in from the options... */
          l7_get_bootp_option( received, TAG_DHCP_SERVER_ID,
                            &received->bp_siaddr.s_addr, sizeof(received->bp_siaddr.s_addr) );
        }

        /* see if it was a DHCP reply or a bootp reply; it could be
                 either.                                                  */
        if ( l7_get_bootp_option( received, TAG_DHCP_MESS_TYPE, &msgtype, sizeof(msgtype)) )
        {
          if ( DHCPOFFER == msgtype )
          {
             /* Get subnet mask */
             if (l7_get_bootp_option( received, TAG_SUBNET_MASK, &ifNetMask, sizeof(ifNetMask)) != L7_TRUE)
                 ifNetMask = osapiHtonl(0xFFFFFF00); /* At least it's not uninitialized */
             /*
             ** Check for IP Address conflicts with other interfaces
             */
             dhcpConfigSemGet ();
             if (sysapiIPConfigConflict((strcmp(bspapiServicePortNameGet(), intf) ? 
                                        FD_CNFGR_NIM_MIN_CPU_INTF_NUM : FD_CNFGR_NIM_MIN_SERV_PORT_INTF_NUM),
                                        osapiNtohl(received->bp_yiaddr.s_addr),
                                        ifNetMask, 
                                        (L7_uint32 *)&errorMsg) == L7_FALSE)
             { 
               /* No conflict */

               /* all is well Save the good packet in *xmit  */
               bcopy( (L7_char8*)received, (L7_char8*)xmit, dhcp_size(received) );
               /* we like the packet, so reset the timeout for next time*/
               reset_timeout( &tv, &timeout_scratch );
               /* Grab T1, T2, lease time from this packet if specified */
               get_lease_params( received, lease );
               *pstate = DHCPSTATE_REQUESTING;
             }
             else /* IP Addr conflict */
             {
               /* Log one error per request */
               if (logMsg == L7_NULL)
               {
                 logMsg++;
                 LOG_MSG("DHCP: Conflict w/offered IP Addr, error=%u\r\n", errorMsg);
               }
             }
             dhcpConfigSemFree ();
          }
        }
        else 
        {
          /* No TAG_DHCP_MESS_TYPE entry so it's a bootp reply*/
          seen_bootp_reply = 1; /* (keep the bootp packet in received)*/
        }

        /* If none of the above state changes occurred, we got a packet
           that "should not happen", OR we have a bootp reply in our
           hand; so listen again with the same timeout, without
           retrying the send, in the hope of getting a DHCP reply.*/
        break;

      case DHCPSTATE_REQUESTING:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP REQUESTING state\n");
        }

        /* Just send what you got with a DHCPREQUEST in the message type.
           then wait for an ACK in DHCPSTATE_REQUEST_RECV.
  
           Fill in the BOOTP request - DHCPREQUEST packet*/
        xmit->bp_op = BOOTREQUEST;
        xmit->bp_flags = osapiHtons(0x8000); /* BROADCAST FLAG*/

        set_fixed_tag( xmit, TAG_DHCP_MESS_TYPE, DHCPREQUEST, 1 );
        xmit->bp_giaddr.s_addr = 0x0;
        set_fixed_tag( xmit, TAG_DHCP_MAX_MSGSZ, BP_MINPKTSZ, 2 );


        /* And this will be a new one:       */
        set_fixed_tag( xmit, TAG_DHCP_REQ_IP, osapiNtohl(xmit->bp_yiaddr.s_addr), 4 );
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_DOMAIN_SERVER);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_GATEWAY);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_SUBNET_MASK);

        local_bp_yiaddr.s_addr = xmit->bp_yiaddr.s_addr;
        l7_get_bootp_option(xmit, TAG_DHCP_SERVER_ID, &(local_bp_siaddr.s_addr), sizeof(local_bp_siaddr.s_addr));

        /* Per RFC 2131 Section 4.4.1 These fields should be initilized to 0 in the   **
        ** BOOTREQUEST message.    Local copies are made for validating the ACK.      */
        xmit->bp_ciaddr.s_addr = 0;
        xmit->bp_yiaddr.s_addr = 0;
        xmit->bp_siaddr.s_addr = 0;
        xmit->bp_giaddr.s_addr = 0;

        if (dhcpcDebugMsgLevel > 0)
        {
          printf( "---------DHCPSTATE_REQUESTING sending:\n" );
          show_bootp( intf, xmit );
        }
        
        if ( osapiSocketSendto(s, (L7_char8 *)xmit, dhcp_size(xmit), 0, 
                    (L7_sockaddr_t *)&broadcast_addr, sizeof(broadcast_addr),&bsent)  != L7_SUCCESS )

        {
          *pstate = DHCPSTATE_FAILED;
          break;
        }


        *pstate = DHCPSTATE_REQUEST_RECV;
        break;

      case DHCPSTATE_REQUEST_RECV:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP REQUEST RECV state\n");
        }

        /* wait for an ACK or a NACK - retry by going back to
           DHCPSTATE_REQUESTING; NACK means go back to INIT.*/

        addrlen = sizeof(rx_addr);

        if ( osapiSelect(maxFdLength, &pReadFd, NULL, NULL, tv.tv_sec, tv.tv_usec) <= 0 )
        {
          if ( ! next_timeout( &tv, &timeout_scratch ) )
          {
            *pstate = DHCPSTATE_FAILED;
            break;
          }
          *pstate = DHCPSTATE_REQUESTING;
          break;
        }

        if ( osapiSocketRecvfrom(s,(L7_char8*) received, sizeof(struct bootp), MSG_DONTWAIT,
                      ( L7_sockaddr_t *)&rx_addr, &addrlen, &brecvd)  != L7_SUCCESS )

        {
          /* No packet arrived
             go to the next larger timeout and re-send:*/
          if ( ! next_timeout( &tv, &timeout_scratch ) )
          {
            *pstate = DHCPSTATE_FAILED;
            break;
          }
          *pstate = DHCPSTATE_REQUESTING;
          break;
        }

        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(rx_addr.sin_addr.s_addr), addrStr);
          printf( "---------DHCPSTATE_REQUEST_RECV received:\n" );
          printf( "...rx_addr is family %d, addr %s, port %d\n",
                  rx_addr.sin_family,
                  addrStr,
                  osapiNtohs(rx_addr.sin_port) );
          show_bootp( intf, received );
        }
        
        if ( received->bp_xid != osapiHtonl(xid) )
        {   /* not the same transaction;*/
          if ( requestCounter < 4 )
            requestCounter++;
          else
            *pstate = DHCPSTATE_FAILED;
          break;                  /* listen again...*/
        }

        /* Cache the server ip from whom the ACK packet is received.
         * This IP would be used later while sending unicast RENEW or RELEASE */
        l7_get_bootp_option( received, TAG_DHCP_SERVER_ID, &(rx_local.server_ip.s_addr), sizeof(rx_local.server_ip.s_addr));

        /* check it was a DHCP reply*/
        if ( l7_get_bootp_option( received, TAG_DHCP_MESS_TYPE, &msgtype, sizeof(msgtype)) )
        {
          if ( DHCPACK == msgtype /* Same offer & server? */
               && received->bp_yiaddr.s_addr == local_bp_yiaddr.s_addr
               && (rx_local.server_ip.s_addr /* ACK sender */ ==
                local_bp_siaddr.s_addr /* OFFER sender */))
          {
    
            struct bootp_serverinfo *xmit_serverinfo = (struct bootp_serverinfo *)xmit;

            /* Save the good packet in *xmit*/
            bcopy( (L7_char8*)received,(L7_char8*) xmit, dhcp_size(received) );

            /* cache the server's ip (siaddr is different) */
            xmit_serverinfo->server_ip.s_addr = rx_local.server_ip.s_addr;

            /* we like the packet, so reset the timeout for next time*/
            reset_timeout( &tv, &timeout_scratch );
            /* Record the new lease and set up timers &c*/
            get_lease_params( received, lease );
            new_lease( lease );

            dhcpClientDnsOptionsAdd(received);

            *pstate = DHCPSTATE_BOUND;
            break;
          }
          if ( DHCPNAK == msgtype /* Same server?*/
               && (rx_local.server_ip.s_addr /* OFFER sender */ ==
                local_bp_siaddr.s_addr /* NAK sender */))
          {
            /* we're bounced!*/
            *pstate = DHCPSTATE_INIT;  /* So back the start of the rigmarole.*/
            reset_timeout( &tv, &timeout_scratch );
            break;
          }
          /* otherwise it's something else, maybe another offer, or a bogus
             NAK from someone we are not asking!
             Just listen again, which implicitly discards it.*/
        }
        break;

      case DHCPSTATE_BOUND:
        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(xmit->bp_yiaddr.s_addr), addrStr);
          printf("\nDHCP BOUND state: addr %s\n", addrStr);
        }

        /* We are happy now, we have our address.*/
        /* So save this address for requesting the same IP address again*/
        previaddr.s_addr = xmit->bp_yiaddr.s_addr;
#ifdef _L7_OS_LINUX_
        /*delete route entry for 0.0.0.0 IP address */
        osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
        /* All done with socket*/
        osapiSocketClose(s);
        *pConfigState= CONFIG_SET;
        return L7_TRUE;


      case DHCPSTATE_RENEWING:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP RENEWING state\n");
        }

        /* Just send what you got with a DHCPREQUEST in the message
           type UNICAST straight to the server.  Then wait for an ACK.*/

        /* Fill in the BOOTP request - DHCPREQUEST packet*/
        xmit->bp_op = BOOTREQUEST;
        xmit->bp_flags = osapiHtons(0); /* No BROADCAST FLAG*/
        /* Use the *client* address here:*/
        xmit->bp_ciaddr.s_addr = xmit->bp_yiaddr.s_addr;

        set_fixed_tag( xmit, TAG_DHCP_MESS_TYPE, DHCPREQUEST, 1 );
        set_fixed_tag( xmit, TAG_DHCP_MAX_MSGSZ, BP_MINPKTSZ, 2 );
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_DOMAIN_SERVER);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_GATEWAY);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_SUBNET_MASK);
        /*as per RFC, these fields MUST NOT be present in DHCP REQUEST*/
        remove_tag( xmit, TAG_DHCP_SERVER_ID);
        remove_tag( xmit, TAG_DHCP_REQ_IP);

         { 
           struct bootp_serverinfo *xmit_serverinfo = (struct bootp_serverinfo *)res;
           /* Set unicast address to *server* */
           server_addr.sin_addr.s_addr = xmit_serverinfo->server_ip.s_addr;
         }

        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(server_addr.sin_addr.s_addr), addrStr);
          printf( "---------DHCPSTATE_RENEWING sending:\n" );
          printf( "UNICAST to family %d, addr %s, port %d\n",
                  server_addr.sin_family,
                  addrStr,
                  osapiNtohs(server_addr.sin_port) );
          show_bootp( intf, xmit );
        }        

        /* UNICAST address of the server:*/
        if ( osapiSocketSendto(s, (L7_char8 *)xmit, dhcp_size(xmit), 0, 
                    (L7_sockaddr_t *)&server_addr, sizeof(server_addr),&bsent)  != L7_SUCCESS )

        {
          *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
          /*delete route entry for 0.0.0.0 IP address */
          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
          osapiSocketClose (s);
          return L7_TRUE;
        }

        *pstate = DHCPSTATE_RENEW_RECV;
        break;

      case DHCPSTATE_RENEW_RECV:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP RENEW RECV state\n");
        }

        /* wait for an ACK or a NACK - retry by going back to
           DHCPSTATE_RENEWING; NACK means go to NOTBOUND. */

        addrlen = sizeof(rx_addr);

        tv.tv_sec = ((lease->t2) > 50000) ? 25 : ((lease->t2)/1000)/2;
        tv.tv_usec = 0;
        
        if (dhcpcDebugMsgLevel > 0)
        {
          printf( "---------DHCPSTATE_RENEW_RECV setup:\n" );
          printf( "wait time is %ld\n", tv.tv_sec);
        }
    
        if ( osapiSelect(maxFdLength, &pReadFd, NULL, NULL, tv.tv_sec, tv.tv_usec) <= 0 )
        {
          reset_timeout( &tv, &timeout_scratch );
          *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
          /*delete route entry for 0.0.0.0 IP address */
          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
          osapiSocketClose (s);

          /* Remove the old DNS Server and Domain Name information in DNS Client */
          dhcpClientDnsOptionsRemove(xmit);

          return L7_TRUE;
        }

        if ( osapiSocketRecvfrom(s,(L7_char8*) received, sizeof(struct bootp), MSG_DONTWAIT,
                      ( L7_sockaddr_t *)&rx_addr, &addrlen, &brecvd)  != L7_SUCCESS )

        {
          /* No packet arrived wait for next time out */
          reset_timeout( &tv, &timeout_scratch );
          *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
          /*delete route entry for 0.0.0.0 IP address */
          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
          osapiSocketClose (s);

          /* Remove the old DNS Server and Domain Name information in DNS Client */
          dhcpClientDnsOptionsRemove(xmit);

          return L7_TRUE;
        }

        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(rx_addr.sin_addr.s_addr), addrStr);
          printf( "---------DHCPSTATE_RENEW_RECV received:\n" );
          printf( "...rx_addr is family %d, addr %s, port %d\n",
                  rx_addr.sin_family,
                  addrStr,
                  osapiNtohs(rx_addr.sin_port) );
          show_bootp( intf, received );
        }
        
        if ( received->bp_xid != osapiHtonl(xid) )
        {   /* not the same transaction;*/
          if ( renewCounter < 4 )
            renewCounter++;     /* listen again... */
          else
          {
            reset_timeout( &tv, &timeout_scratch );
            *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
            /*delete route entry for 0.0.0.0 IP address */
            osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
            osapiSocketClose (s);
            return L7_TRUE;
          }
          break;                  
        }

        if ( 0 == received->bp_siaddr.s_addr )
        {
          /* then fill in from the options... */
          l7_get_bootp_option( received, TAG_DHCP_SERVER_ID,
                            &received->bp_siaddr.s_addr, sizeof(received->bp_siaddr.s_addr) );
        }

        /* check it was a DHCP reply*/
        if ( l7_get_bootp_option( received, TAG_DHCP_MESS_TYPE, &msgtype, sizeof(msgtype)) )
        {
          if ( DHCPACK == msgtype  /* Same offer? */
               && received->bp_yiaddr.s_addr == xmit->bp_yiaddr.s_addr )
          {
            /* Remove the old DNS Server and Domain Name information in DNS Client */
            dhcpClientDnsOptionsRemove(xmit);

            /* Save the good packet in *xmit */
            bcopy( (L7_uchar8*)received, (L7_uchar8*)xmit, dhcp_size(received) );
            /* we like the packet, so reset the timeout for next time */
            reset_timeout( &tv, &timeout_scratch );
            /* Record the new lease and set up timers &c */
            get_lease_params( received, lease );
            new_lease( lease );

            /* Add the DNS Server and Domain Name information in DNS Client */
            dhcpClientDnsOptionsAdd(received);

            *pstate = DHCPSTATE_BOUND;
            break;
          }
          if ( DHCPNAK == msgtype )
          { /* we're bounced! */
            *pstate = DHCPSTATE_NOTBOUND;  /* So quit out. */
            reset_timeout( &tv, &timeout_scratch );
            break;
          }
          /* otherwise it's something else, maybe another offer.
             Just listen again, which implicitly discards it.*/
        }
        break;

      case DHCPSTATE_REBINDING:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP REBINDING state\n");
        }

        /* Just send what you got with a DHCPREQUEST in the message type.
           Then wait for an ACK.  This one is BROADCAST.
  
           Fill in the BOOTP request - DHCPREQUEST packet */
        xmit->bp_op = BOOTREQUEST;
        xmit->bp_flags = osapiHtons(0); /* no BROADCAST FLAG*/
        /* Use the *client* address here: */
        xmit->bp_ciaddr.s_addr = xmit->bp_yiaddr.s_addr;

        set_fixed_tag( xmit, TAG_DHCP_MESS_TYPE, DHCPREQUEST, 1 );
        set_fixed_tag( xmit, TAG_DHCP_MAX_MSGSZ, BP_MINPKTSZ, 2 );
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_DOMAIN_SERVER);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_GATEWAY);
        add_fixed_tag( xmit, TAG_DHCP_PARM_REQ_LIST, TAG_SUBNET_MASK);
        /*as per RFC, these fields MUST NOT be present in DHCP REQUEST*/
        remove_tag( xmit, TAG_DHCP_SERVER_ID);
        remove_tag( xmit, TAG_DHCP_REQ_IP);

        if (dhcpcDebugMsgLevel > 0)
        {
          printf( "---------DHCPSTATE_REBINDING sending:\n" );
          printf( intf, xmit );
        }
   
        if ( osapiSocketSendto(s, (L7_char8 *)xmit, dhcp_size(xmit), 0, 
                    (L7_sockaddr_t *)&broadcast_addr, sizeof(broadcast_addr),&bsent)  != L7_SUCCESS )
        {
          *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
          /*delete route entry for 0.0.0.0 IP address */
          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
          osapiSocketClose (s);
          return L7_TRUE;
        }

        *pstate = DHCPSTATE_REBIND_RECV;
        break;

      case DHCPSTATE_REBIND_RECV:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP REBIND RECV state\n");
        }

        /* wait for an ACK or a NACK - retry by going back to
           DHCPSTATE_REBINDING; NACK means go to NOTBOUND. */

        addrlen = sizeof(rx_addr);

        tv.tv_sec = ((lease->expiry) > 50000) ? 25 : ((lease->expiry)/1000)/2;
        tv.tv_usec = 0;
        
        if ( osapiSelect(maxFdLength, &pReadFd, NULL, NULL, tv.tv_sec, tv.tv_usec) <= 0 )
        {
          reset_timeout( &tv, &timeout_scratch );
          *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
          /*delete route entry for 0.0.0.0 IP address */
          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
          osapiSocketClose (s);
          return L7_TRUE;
        }

        if ( osapiSocketRecvfrom(s,(L7_char8*) received, sizeof(struct bootp), MSG_DONTWAIT,
                      ( L7_sockaddr_t *)&rx_addr, &addrlen, &brecvd)  != L7_SUCCESS )

        {
          /* No packet arrived wait for next time out */
          reset_timeout( &tv, &timeout_scratch );
          *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
          /*delete route entry for 0.0.0.0 IP address */
          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
          osapiSocketClose (s);
          return L7_TRUE;
        }

        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(rx_addr.sin_addr.s_addr), addrStr);
          printf( "---------DHCPSTATE_REBIND_RECV received:\n" );
          printf( "...rx_addr is family %d, addr %s, port %d\n",
                  rx_addr.sin_family,
                  addrStr,
                  osapiNtohs(rx_addr.sin_port) );
          show_bootp( intf, received );
        }
        
        if ( received->bp_xid != osapiHtonl(xid) )
        {   /* not the same transaction;     */
          if ( rebindCounter < 4 )
            rebindCounter++;      /* listen again... */
          else
          {
            reset_timeout( &tv, &timeout_scratch );
            *pstate = DHCPSTATE_BOUND;
#ifdef _L7_OS_LINUX_
            /*delete route entry for 0.0.0.0 IP address */
            osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
            osapiSocketClose (s);
            return L7_TRUE;
          }
          break;                  
        }

        if ( 0 == received->bp_siaddr.s_addr )
        {
          /* then fill in from the options...    */
          l7_get_bootp_option( received, TAG_DHCP_SERVER_ID,
                            &received->bp_siaddr.s_addr,  sizeof(received->bp_siaddr.s_addr));
        }

        /* check it was a DHCP reply */
        if ( l7_get_bootp_option( received, TAG_DHCP_MESS_TYPE, &msgtype, sizeof(msgtype)) )
        {
          if ( DHCPACK == msgtype  /* Same offer?*/
               && received->bp_yiaddr.s_addr == xmit->bp_yiaddr.s_addr )
          {
            /* Remove the old DNS Server and Domain Name information in DNS Client */
            dhcpClientDnsOptionsRemove(xmit);

            /* Save the good packet in *xmit   */
            bcopy((char *)received, (L7_uchar8*)xmit, dhcp_size(received) );
            /* we like the packet, so reset the timeout for next time */
            reset_timeout( &tv, &timeout_scratch );
            /* Record the new lease and set up timers &c  */
            get_lease_params( received, lease );
            new_lease( lease );

            /* Add the DNS Server and Domain Name information in DNS Client */
            dhcpClientDnsOptionsAdd(received);

            *pstate = DHCPSTATE_BOUND;
            *pConfigState= CONFIG_SET;
            break;
          }
          else if ( DHCPNAK == msgtype )
          { /* we're bounced! */
            *pstate = DHCPSTATE_NOTBOUND;  /* So back the start of the rigmarole.*/
            reset_timeout( &tv, &timeout_scratch );
            break;
          }
          /* otherwise it's something else, maybe another offer.
             Just listen again, which implicitly discards it.*/
        }
        break;

      case DHCPSTATE_BOOTP_FALLBACK:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP BOOTP FALLBACK state\n");
        }

        /* All done with socket*/
        osapiSocketClose(s);
        /* And no lease should have become active, but JIC*/
        no_lease( lease );
        *pstate = DHCPSTATE_FAILED;
        break;


      case DHCPSTATE_NOTBOUND:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP NOTBOUND state\n");
        }

        /* All done with socket*/
        osapiSocketClose(s);
        /* No lease active        */
        no_lease( lease );
        /* Bring down the interface*/
        *pstate = DHCPSTATE_FAILED;
        break;


      case DHCPSTATE_FAILED:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP FAILED state\n");
        }

        /* All done with socket */
        osapiSocketClose(s);
        /* No lease active*/
        no_lease( lease );

		dhcpConfigSemGet ();
        if ( strcmp(bspapiServicePortNameGet(), intf) == 0 )
        {
          servPortConfigMode = simGetServPortConfigMode();
          if ( servPortConfigMode != L7_SYSCONFIG_MODE_DHCP )
		  {
			  dhcpConfigSemFree ();
#ifdef _L7_OS_LINUX_ 
                          /*delete route etry for 0.0.0.0 IP address */
                          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
			  return L7_FALSE;
		  }

          currentIpAddr = simGetServPortIPAddr();
          if ( currentIpAddr != L7_NULL )
          {
              simSetServPortIPAddr(0,L7_FALSE);
          }
          simSetServPortIPNetMask(L7_NULL, L7_FALSE);
          simSetServPortIPGateway(L7_NULL, L7_FALSE);
        }
        else if ( strcmp(L7_DTL_PORT_IF, intf) == 0 )
        {
          networkConfigMode = simGetSystemConfigMode();
          if ( networkConfigMode != L7_SYSCONFIG_MODE_DHCP )
		  {
			  dhcpConfigSemFree ();
#ifdef _L7_OS_LINUX_
                          /*delete route etry for 0.0.0.0 IP address */
                          osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
			  return L7_FALSE;
		  }

          currentIpAddr = simGetSystemIPAddr();
           
          if ( currentIpAddr != FD_SIM_DEFAULT_NETWORK_IP_ADDR_NETGEAR)
          {
              simSetSystemIPAddr(FD_SIM_DEFAULT_NETWORK_IP_ADDR_NETGEAR,L7_FALSE);
          }
          simSetSystemIPNetMask(FD_SIM_DEFAULT_NETWORK_SUBNET_MASK_NETGEAR);
          simSetSystemIPGateway(FD_SIM_DEFAULT_NETWORK_GATEWAY_ADDR_NETGEAR, L7_FALSE);

        }
		dhcpConfigSemFree ();

        /* Unconditionally down the interface.*/
        do_dhcp_down_net( intf, pstate);
#ifdef _L7_OS_LINUX_
        /*delete route etry for 0.0.0.0 IP address */
        osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
        return L7_FALSE;

      case DHCPSTATE_DO_RELEASE:
        if (dhcpcDebugMsgLevel > 0)
        {
          printf("\nDHCP DO RELEASE state\n");
        }

        /* We have been forced here by external means, to release the
           lease for graceful shutdown.
  
           Just send what you got with a DHCPRELEASE in the message
           type UNICAST straight to the server.  No ACK.  Then go to
           NOTBOUND state.*/

        xmit->bp_op = BOOTREQUEST;
        xmit->bp_flags = osapiHtons(0); /* no BROADCAST FLAG*/
        /* Use the *client* address here: */
        xmit->bp_ciaddr.s_addr = xmit->bp_yiaddr.s_addr;

        set_fixed_tag( xmit, TAG_DHCP_MESS_TYPE, DHCPRELEASE, 1 );
       { 
         struct bootp_serverinfo *xmit_serverinfo = (struct bootp_serverinfo *)res;
         /* Set unicast address to *server* */
         server_addr.sin_addr.s_addr = xmit_serverinfo->server_ip.s_addr;
       }  
        if (dhcpcDebugMsgLevel > 0)
        {
          memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
          osapiInetNtoa(osapiNtohl(server_addr.sin_addr.s_addr), addrStr);
          printf( "---------DHCPSTATE_DO_RELEASE sending:\n" );
          printf( "UNICAST to family %d, addr %s, port %d\n",
                  server_addr.sin_family,
                  addrStr,
                  osapiNtohs(server_addr.sin_port) );
          show_bootp( intf, xmit );                                 
        }

                    /* UNICAST address of the server:*/
        if ( osapiSocketSendto(s, (L7_char8 *)xmit, dhcp_size(xmit), 0, 
                    (L7_sockaddr_t *)&server_addr, sizeof(server_addr),&bsent)  != L7_SUCCESS )

        {
          *pstate = DHCPSTATE_FAILED;
          break;
        }
              
        /* Remove the old DNS Server and Domain Name information in DNS Client */
        dhcpClientDnsOptionsRemove(xmit);

        *pstate = DHCPSTATE_NOTBOUND;
        break;

      default:
        LOG_MSG("DHCP default state -- should not be here\n");
        no_lease( lease );
#ifdef _L7_OS_LINUX_
        /*delete route etry for 0.0.0.0 IP address */
        osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
        osapiSocketClose(s);
        return L7_FALSE;
    }
  }
  /* NOTREACHED */
#ifdef _L7_OS_LINUX_
  /*delete route etry for 0.0.0.0 IP address */
  osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff, 0, 0, 1);
#endif
  osapiSocketClose (s);
  return L7_FALSE;
}


/* ------------------------------------------------------------------------
   Bring an interface down, failed to initialize it or lease is expired
   Also part of normal startup, bring down for proper reinitialization*/


/**************************************************************************
* @purpose  Bring down the interface by deleting IP interface 
*
* @param    L7_char8*      Interface Name
* @param    L7_short16*    Ptr to DHCP state
*
* @returns  L7_TRUE or L7_FALSE
*
* @comments None.
*
* @end
*************************************************************************/
L7_int32
do_dhcp_down_net(const L7_char8 *intf, L7_short16 *pstate)
{
  L7_sockaddr_in_t *addrp;
  struct ifreq ifr;
  L7_char8   addrStr[OSAPI_INET_NTOA_BUF_SIZE];
  L7_char8   ifname[IFNAMSIZ];
  L7_sockaddr_in_t *addrInt;
  L7_int32   s;

  if (osapiSocketCreate(L7_AF_INET, L7_SOCK_DGRAM, 0, &s) == L7_FAILURE)
  {
    LOG_MSG("do_dhcp_down_net(): socket creation failed");
    return L7_FALSE;
  }

  addrp = (L7_sockaddr_in_t *) &ifr.ifr_addr;

  sprintf(ifname,"%s%s",intf, "0");
  addrInt= (L7_sockaddr_in_t *)(&ifr.ifr_addr);

  /* Remove any existing address*/
  if ( DHCPSTATE_FAILED   == *pstate
       || DHCPSTATE_INIT    == *pstate
       || DHCPSTATE_NOTBOUND  == *pstate
       || 0         == *pstate )
  {
    /* it was configured for broadcast only, "half-up"*/
    memset(addrp, 0, sizeof(*addrp));
    addrp->sin_family = AF_INET;

#ifdef _L7_OS_VXWORKS_
    addrp->sin_len = sizeof(*addrp);
#endif /* _L7_OS_VXWORKS_ */
    addrp->sin_port = 0;
    addrp->sin_addr.s_addr = osapiHtonl(INADDR_ANY);
  }
  else
  {
    /* get the specific address that was used*/
    if ( osapiNetIPGetAddr(ifr.ifr_name, (L7_uint32 *)&addrInt->sin_addr.s_addr) != L7_SUCCESS )
    {
      perror("SIOCGIFADDR 1");
      osapiSocketClose (s);
      return L7_FALSE;
    }
  }

  /*Delete route entries for this interface */

  addrInt= (L7_sockaddr_in_t *)(&ifr.ifr_addr);

  memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
  osapiInetNtoa(osapiNtohl(addrInt->sin_addr.s_addr), addrStr);

  /* All done with socket*/
  osapiSocketClose(s);

  return L7_TRUE;
}


/**************************************************************************
* @purpose  Send DHCPRELEASE to dhcp server 
*
* @param    L7_char8*      Interface Name
* @param    struct bootp*   Ptr to bootp structure
* @param    L7_short16*    Ptr to DHCP state
* @param    struct dhcp_lease*  Ptr to lease structure
*
* @returns  L7_TRUE or L7_FALSE
*
* @comments None.
*
* @end
*************************************************************************/
L7_int32
do_dhcp_release(const L7_char8 *intf, struct bootp *res,
                L7_short16 *pstate, struct dhcp_lease *lease)
{
  L7_short16 configState;
  configState=CONFIG_NONE;
  if ( 0                           != *pstate
       && DHCPSTATE_INIT           != *pstate
       && DHCPSTATE_NOTBOUND       != *pstate
       && DHCPSTATE_FAILED         != *pstate
       && DHCPSTATE_BOOTP_FALLBACK != *pstate )
  {
    *pstate = DHCPSTATE_DO_RELEASE;
    do_dhcp( intf, res, pstate, lease, &configState); /* to send the release packet*/
    osapiSleep( 1 );             /* to let it leave the building  */
  }
  return L7_TRUE;
}


/* Following functions have been copied from Redhat/eCos bootp_support.c */

static L7_char8 *_bootp_op[] = {"", "REQUEST", "REPLY"};
static L7_char8 *_bootp_hw_type[] = {"", "Ethernet", "Exp Ethernet", "AX25",
  "Pronet", "Chaos", "IEEE802", "Arcnet"};

static L7_char8 *_dhcpmsgs[] = {"","DISCOVER", "OFFER", "REQUEST", "DECLINE",
  "ACK", "NAK", "RELEASE"};


/**************************************************************************
* @purpose  To show bootp parameters
*
* @param    L7_char8*      Interface Name
* @param    L7_short16*    Ptr to bootp structure
*
* @returns  Void
*
* @comments None.
*
* @end
*************************************************************************/
void
show_bootp(const L7_char8 *intf, struct bootp *bp)
{
  L7_int32  i, len;
  L7_uchar8 *op, *ap = 0, optover;
  L7_uchar8 name[128];
  struct    in_addr addr[32];
  L7_char8  addrStr[OSAPI_INET_NTOA_BUF_SIZE];

  printf("BOOTP[%s] op: %s\n", intf, _bootp_op[bp->bp_op]);
  printf("       htype: %s\n", _bootp_hw_type[bp->bp_htype]);
  printf("        hlen: %d\n", bp->bp_hlen );
  printf("        hops: %d\n", bp->bp_hops );
  printf("         xid: 0x%lx\n", osapiNtohl(bp->bp_xid) );
  printf("        secs: %d\n", bp->bp_secs );
  printf("       flags: 0x%x\n", bp->bp_flags );
  printf("       hw_addr: ");
  for ( i = 0;  i < bp->bp_hlen;  i++ )
  {
    printf("%02x", bp->bp_chaddr[i]);
    if ( i != (bp->bp_hlen-1) ) printf(":");
  }
  printf("\n");

  memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
  osapiInetNtoa(osapiNtohl(bp->bp_ciaddr.s_addr), addrStr);
  printf("     client IP: %s\n", addrStr);

  memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
  osapiInetNtoa(osapiNtohl(bp->bp_yiaddr.s_addr), addrStr);
  printf("         my IP: %s\n", addrStr);

  memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
  osapiInetNtoa(osapiNtohl(bp->bp_siaddr.s_addr), addrStr);
  printf("     server IP: %s\n", addrStr);

  memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
  osapiInetNtoa(osapiNtohl(bp->bp_giaddr.s_addr), addrStr);
  printf("    gateway IP: %s\n", addrStr);

  optover = 0; /* See whether sname and file are overridden for options*/
  (void)l7_get_bootp_option( bp, TAG_DHCP_OPTOVER, &optover, sizeof(optover));
  if ( !(1 & optover) && bp->bp_sname[0] )
    printf("        server: %s\n", bp->bp_sname);
  if ( ! (2 & optover) && bp->bp_file[0] )
    printf("          file: %s\n", bp->bp_file);
  if ( bp->bp_vend[0] )
  {
    printf("  options:\n");
    op = &bp->bp_vend[4];
    while ( *op != TAG_END )
    {
      switch ( *op )
      {
        case TAG_SUBNET_MASK:
        case TAG_GATEWAY:
        case TAG_IP_BROADCAST:
        case TAG_DOMAIN_SERVER:
          ap = (L7_uchar8 *)&addr[0];
          len = *(op+1);
          for ( i = 0;  i < len;  i++ )
          {
            *ap++ = *(op+i+2);
          }
          if ( *op == TAG_SUBNET_MASK )   ap =  "  subnet mask";
          if ( *op == TAG_GATEWAY )       ap =  "      gateway";
          if ( *op == TAG_IP_BROADCAST )  ap =  " IP broadcast";
          if ( *op == TAG_DOMAIN_SERVER ) ap =  "domain server";
          printf("      %s: ", ap);
          ap = (L7_uchar8 *)&addr[0];
          while ( len > 0 )
          {
            memset(addrStr, 0, OSAPI_INET_NTOA_BUF_SIZE);
            osapiInetNtoa(osapiNtohl((L7_uint32)(((struct in_addr *)ap)->s_addr)),
                          addrStr);
            printf("%s", addrStr);
            len -= sizeof(struct in_addr);
            ap += sizeof(struct in_addr);
            if ( len ) printf(", ");
          }
          printf("\n");
          break;
        case TAG_DOMAIN_NAME:
        case TAG_HOST_NAME:
          for ( i = 0;  i < *(op+1);  i++ )
          {
            name[i] = *(op+i+2);
          }
          name[*(op+1)] = '\0';
          if ( *op == TAG_DOMAIN_NAME ) ap =  " domain name";
          if ( *op == TAG_HOST_NAME )   ap =  "   host name";
          printf("       %s: %s\n", ap, name);
          break;
        case TAG_DHCP_MESS_TYPE:
          printf("        DHCP message: %d %s\n",
                  op[2], _dhcpmsgs[op[2]] );
          break;
        case TAG_DHCP_REQ_IP:
          printf("        DHCP requested ip: %d.%d.%d.%d\n",
                  op[2], op[3], op[4], op[5] );  
          break;
        case TAG_DHCP_LEASE_TIME   :
        case TAG_DHCP_RENEWAL_TIME :
        case TAG_DHCP_REBIND_TIME  :
          printf("        DHCP time %d: %d\n",
                  *op, ((((((op[2]<<8)+op[3])<<8)+op[4])<<8)+op[5]) );

          break;
        case TAG_DHCP_SERVER_ID    :
          printf("        DHCP server id: %d.%d.%d.%d\n",
                  op[2], op[3], op[4], op[5] );  
          break;

        case TAG_DHCP_OPTOVER      :
        case TAG_DHCP_PARM_REQ_LIST:
        case TAG_DHCP_TEXT_MESSAGE :
        case TAG_DHCP_MAX_MSGSZ    :
        case TAG_DHCP_CLASSID      :
        case TAG_DHCP_CLIENTID     :
          printf("        DHCP option: %x/%d.%d:", *op, *op, *(op+1));
          if ( 1 == op[1] )
          {
            printf( " %d", op[2] );
          }
          else if ( 2 == op[1] )
          {
            printf( " %d", (op[2]<<8) + op[3] );
          }
          else if ( 4 == op[1] )
          {
            printf( " %d", ((((((op[2]<<8)+op[3])<<8)+op[4])<<8)+op[5]) );
          }
          else
          {
            for ( i = 2; i < 2 + op[1]; i++ )
              printf(" %d",op[i]);
          }
          printf("\n");
          break;

        default:
          printf("        Unknown option: %x/%d.%d:", *op, *op, *(op+1));
          for ( i = 2; i < 2 + op[1]; i++ )
            printf(" %d",op[i]);
          printf("\n");
          break;
      }                
      /*temp op += *(op+1)+2; */
      op += *(op+1)+2;
    }
  }
}

/**************************************************************************
* @purpose  Bring down the interface by deleting IP interface 
*
* @param    struct bootp*      Ptr to bootp structure
* @param    L7_uchar8*         Tag Identity
* @param    Void*              Option Value
* @param    L7_int32*          Option Length
*
* @returns  L7_TRUE or L7_FALSE
*
* @comments None.
*
* @end
*************************************************************************/
L7_short16
l7_get_bootp_option(struct bootp *bp, L7_uchar8 tag, void *opt, L7_int32 optLength)
{
  L7_uchar8 *val = (L7_uchar8 *)opt;
  L7_int32 i;
  L7_ushort16 optover;
  L7_uchar8 * op;
  L7_uchar8 *bufend;

  op= &bp->bp_vend[4];
  bufend=op+sizeof(bp->bp_vend)-4; /* set the options buffer boundary */
  while ( *op != TAG_END && op < bufend )
  {
    if ( *op == tag )
    {
      for ( i = 0;  i < *(op+1);  i++ )
      {
        if((L7_int32)val < ((L7_int32)opt+optLength))
        {
          *val++ = *(op+i+2);             
        }
      }                                   
      return L7_TRUE;                        
    }
    op += *(op+1)+2;                        
  }    

  /* Check if we have reached the end of the option boundary */
  if(op >= bufend)
  {
    /* The options in the dhcp packet contains neigher the TAG asked for nor the TAG_END */
    LOG_MSG("l7_get_bootp_option(): TAG %d not found in the buffer\n", tag);
    return L7_FALSE;
  }

  if ( TAG_DHCP_OPTOVER == tag ) /* prevent recursion > once*/
    return L7_FALSE;
  /* else, look for that tag to see if there's more...*/
  optover = 0;
  if ( ! l7_get_bootp_option( bp, TAG_DHCP_OPTOVER, &optover, sizeof(optover)) )
    return L7_FALSE;

#define SCANTAG( ptrM ) {          \
    L7_uchar8 *opM = (ptrM);                  \
    while (*opM != TAG_END) {                    \
        if (*opM == tag) {                       \
            for (i = 0;  i < *(opM+1);  i++) {   \
              if((L7_int32)val < ((L7_int32)opt+optLength)) { \
                *val++ = *(opM+i+2);             \
              }                                 \
            }                                   \
            return L7_TRUE;                        \
        }                                       \
        opM += *(opM+1)+2;                        \
    }                                           \
  }

  if ( 1 & optover ) /* then the file field also holds options*/
    SCANTAG( &bp->bp_file[0] );

  if ( 2 & optover ) /* then the sname field also holds options*/
    SCANTAG( &bp->bp_sname[0] );

  return L7_FALSE;
}                    

/**************************************************************************
* @purpose  To set DHCP client debug level
*
* @param    level   Debug level
* 
*
* @returns  L7_SUCCESS or L7_FAILURE
*
* @comments None.
*
* @end
*************************************************************************/
L7_uint32
dhcpcDebugMsgLvlSet( L7_uint32 level )
{
  dhcpcDebugMsgLevel = level;

  return L7_SUCCESS;
}

/* EOF dhcp_prot.c*/


