/*********************************************************************
*
* (C) Copyright Broadcom Corporation 2000-2007
*
**********************************************************************
* @filename dhcp_support.c
*
* @purpose dhcp main file
*
* @component dhcp
*
* @comments Ported from ecos dhcp_support.c (refer Next Header)
*
* @create 17/05/2001
*
* @author  ported by Paresh Jain
*
* @end
*
**********************************************************************/

/*************************************************************
                    
*************************************************************/



/*==========================================================================

        dhcp_support.c

        Support code == friendly API for DHCP client

  ==========================================================================
  ####COPYRIGHTBEGIN####

   -------------------------------------------
   The contents of this file are subject to the Red Hat eCos Public License
   Version 1.1 (the "License"); you may not use this file except in
   compliance with the License.  You may obtain a copy of the License at
   http:  www.redhat.com/

   Software distributed under the License is distributed on an "AS IS"
   basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the
   License for the specific language governing rights and limitations under
   the License.

   The Original Code is eCos - Embedded Configurable Operating System,
   released September 30, 1998.

   The Initial Developer of the Original Code is Red Hat.
   Portions created by Red Hat are
   Copyright (C) 1998, 1999, 2000 Red Hat, Inc.
   All Rights Reserved.
   -------------------------------------------

  ####COPYRIGHTEND####
  ==========================================================================
  #####DESCRIPTIONBEGIN####

   Author(s):   hmt
   Contributors: gthomas
   Date:        2000-07-01
   Purpose:     DHCP support
   Description:

  ####DESCRIPTIONEND####

  ========================================================================*/

#include <l7_common.h>
#include <l7netapi.h>
#include <datatypes.h>
#include <default_cnfgr.h>
#include <usmdb.h>
#include <osapi.h>
#include <osapi_support.h>
#include <simapi.h>
#include <log.h>
#include <dhcp.h>
#include <dtlapi.h>
#include <bspapi.h>

/*LVL7:eth0 stands for service port and eth1 stands for network port */
struct bootp_serverinfo eth0_bootp_data;
struct bootp_serverinfo eth1_bootp_data;

#ifdef _L7_OS_ECOS_
#include <network.h>
#else
L7_char8   eth0_up;
#endif

L7_char8   eth1_up;


L7_short16   eth0_dhcpstate = 0;
L7_short16   eth1_dhcpstate = 0;


struct dhcp_lease eth0_lease;
struct dhcp_lease eth1_lease;

void* dhcp_needs_attention;
L7_int32 L7_DHCP_task_id;

L7_uint32 servPortConfigStatus = L7_SYSCONFIG_MODE_NOT_COMPLETE;
L7_uint32 networkConfigStatus = L7_SYSCONFIG_MODE_NOT_COMPLETE;

/* Only log changes to DHCP status. */
L7_BOOL servPortLogStatus    = L7_TRUE;
L7_BOOL networkPortLogStatus = L7_TRUE;

L7_short16 servPortDHCPState;
L7_short16 networkDHCPState;

struct in_addr previaddr;

/* DHCP Configuraration protection semaphore.
*/
static void * dhcpConfigSem = 0;
/**************************************************************************
* @purpose  Extract parameters from dhcp parameters structue.
*
* @param    bootp   DHCP structure
* @param    L7_bootp_dhcp_t Network Param Structure
*
*
* @returns  L7_SUCCESS or L7_FAILURE
*
* @comments Reworked for @2975.
*
* @end
*************************************************************************/

L7_int32 getParamsFromDhcpStruct(struct bootp* dhcpParams, L7_bootp_dhcp_t* networkParams)
{
  L7_uint32 netMask;
  L7_uint32 defaultGateway[L7_DHCPS_DEFAULT_ROUTER_MAX];

  /* Get subnet mask & default gw from bootp packet*/
  if (l7_get_bootp_option( dhcpParams, TAG_SUBNET_MASK, &netMask, sizeof(netMask)) != L7_TRUE)
	  netMask = osapiHtonl(0xFFFFFF00); /* At least it's not uninitialized */
  if (l7_get_bootp_option( dhcpParams, TAG_GATEWAY, (void *)defaultGateway, sizeof(defaultGateway)) != L7_TRUE)
	  defaultGateway[0] = 0;

  networkParams->ip=osapiNtohl(dhcpParams->bp_yiaddr.s_addr);
  networkParams->netMask= osapiNtohl(netMask);
  networkParams->gateway= osapiNtohl(defaultGateway[0]);

  /* Verify the validity of the ip address, subnet mask & gateway together */
  return usmDbIpInfoValidate(networkParams->ip, networkParams->netMask, networkParams->gateway);
}


/**************************************************************************
* @purpose  Calls DHCP state machine function for service port config
*     parameters.
*
* @returns  L7_SUCCESS or L7_FAILURE
*
* @comments None.
*
* @end
*************************************************************************/
L7_int32 dhcp_bind_service_port( void )
{
  L7_uint32       servPortConfigMode;

  dhcpConfigSemGet ();
  /* for the service port */
  usmDbAgentBasicConfigServPortConfigProtocolDesiredGet(USMDB_UNIT_CURRENT, &servPortConfigMode);
  dhcpConfigSemFree ();

  /* Run the dhcp state machine for service port if Config Mode and status are true ...*/
  if ( (servPortConfigMode == L7_SYSCONFIG_MODE_DHCP) &&
       (servPortConfigStatus != L7_SYSCONFIG_MODE_COMPLETE) )
  {

    eth0_up = do_dhcp(bspapiServicePortNameGet(), (struct bootp *)&eth0_bootp_data,
                      &eth0_dhcpstate, &eth0_lease, &servPortDHCPState);
  }

  if ( DHCPSTATE_FAILED == eth0_dhcpstate )
  {
    if ( L7_TRUE == servPortLogStatus )
    {
      LOG_MSG("dhcp_bind(): servPort dhcpstate failed \n");
      servPortLogStatus = L7_FALSE;
    }
    return L7_FAILURE;
  }

  return L7_SUCCESS;
}

/**************************************************************************
* @purpose  Calls DHCP state machine function for network port config
*     parameters.
*
* @returns  L7_SUCCESS or L7_FAILURE
*
* @comments None.
*
* @end
*************************************************************************/

L7_int32 dhcp_bind_network( void )
{
  L7_uint32       networkConfigMode;

  /* for inband */
  dhcpConfigSemGet ();
  usmDbAgentBasicConfigNetworkConfigProtocolDesiredGet(USMDB_UNIT_CURRENT, &networkConfigMode);
  dhcpConfigSemFree ();

  /* Run the state machine for network port...*/
  if ( (networkConfigMode == L7_SYSCONFIG_MODE_DHCP) &&
       (networkConfigStatus != L7_SYSCONFIG_MODE_COMPLETE) )
  {

    eth1_up = do_dhcp(L7_DTL_PORT_IF, (struct bootp *)&eth1_bootp_data, &eth1_dhcpstate,
					  &eth1_lease, &networkDHCPState);
  }

  if ( DHCPSTATE_FAILED == eth1_dhcpstate )
  {
    if ( L7_TRUE == networkPortLogStatus )
    {
      LOG_MSG("dhcp_bind(): networkPort dhcpstate failed \n");
      networkPortLogStatus = L7_FALSE;
    }
    return L7_FAILURE;
  }
  return L7_SUCCESS;
}


/**************************************************************************
* @purpose  DHCP task entry poit for configuration of Sercvice port and network
*            port parameters
*
*
* @returns  L7_SUCCESS or L7_FAILURE
*
* @comments None.
*
* @end
*************************************************************************/
L7_RC_t L7_dhcp(void)
{

  L7_short16 old_eth0_dhcpstate = eth0_dhcpstate;
  L7_short16 old_eth1_dhcpstate = eth1_dhcpstate;

  L7_bootp_dhcp_t params;
  L7_uint32       ibMode, ibMode_last = 0;             /* bootp/DHCP or none */
  L7_uint32       spMode, spMode_last = 0;             /* bootp/DHCP or none */
  L7_uint32       ibMode2, spMode2;

  eth0_lease.dhcpState= &eth0_dhcpstate;
  eth1_lease.dhcpState= &eth1_dhcpstate;
  eth0_lease.configStatus= &servPortConfigStatus;
  eth1_lease.configStatus= &networkConfigStatus;
  eth0_lease.net_t1 = eth1_lease.net_t1 = 0;
  eth0_lease.net_t2 = eth1_lease.net_t2 = 0;
  eth0_lease.net_expiry = eth1_lease.net_expiry = 0;

  servPortDHCPState=CONFIG_NONE;
  networkDHCPState= CONFIG_NONE;

  previaddr.s_addr = 0;

#if defined NETGEAR_GS7XXT || defined NETGEAR_GS1XXT
  dhcpFilterSemaTake();
  dhcpFilterSemaDelete();
#endif 

  while ( L7_TRUE )
  {

	dhcpConfigSemGet ();

    usmDbAgentBasicConfigServPortConfigProtocolDesiredGet(USMDB_UNIT_CURRENT, &spMode);
    if (spMode != spMode_last)
    {
      servPortLogStatus = L7_TRUE;
    }
    spMode_last = spMode;

    usmDbAgentBasicConfigNetworkConfigProtocolDesiredGet(USMDB_UNIT_CURRENT, &ibMode);
    if (ibMode != ibMode_last)
    {
      networkPortLogStatus = L7_TRUE;
    }
    ibMode_last = ibMode;

	dhcpConfigSemFree ();

    if ( spMode == L7_SYSCONFIG_MODE_DHCP )
    { /* Service Port not configured for DHCP */
      if ( eth0_dhcpstate != DHCPSTATE_BOUND )
      { /* Not in a bound state */
        if ( dhcp_bind_service_port() == L7_SUCCESS )
        {
          if ( eth0_up &&
               eth0_dhcpstate == DHCPSTATE_BOUND &&
               old_eth0_dhcpstate != eth0_dhcpstate &&
               servPortDHCPState== CONFIG_SET )
          {
		    L7_uchar8 ifname[20];

			sprintf(ifname, "%s0", bspapiServicePortNameGet());

            /* Set new IP parameters */
            if ( L7_SUCCESS == getParamsFromDhcpStruct((struct bootp *)&eth0_bootp_data, &params) )
            {
			  dhcpConfigSemGet ();
              usmDbAgentBasicConfigServPortConfigProtocolDesiredGet(USMDB_UNIT_CURRENT, &spMode2);
              if (spMode2 == L7_SYSCONFIG_MODE_DHCP) { /* Close race condition */
                /*Set IP Addresss, Default Gateway, Netmask */
			    usmDbServicePortIPAddrSet(USMDB_UNIT_CURRENT, params.ip);
			    usmDbServicePortNetMaskSet(USMDB_UNIT_CURRENT, params.netMask);
			    usmDbServicePortGatewaySet(USMDB_UNIT_CURRENT, params.gateway);
              }
			  dhcpConfigSemFree ();

#ifdef _L7_OS_LINUX_    
              /*delete route entry for 0.0.0.0 IP address */
              osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff,
			    0, 0, 1);
#endif
              servPortConfigStatus = L7_SYSCONFIG_MODE_COMPLETE;
              servPortLogStatus = L7_TRUE;
            }
          }
        }

        if ( eth0_dhcpstate == DHCPSTATE_FAILED )
		{
			eth0_dhcpstate= 0;
		}
      }
    }

    if ( ibMode == L7_SYSCONFIG_MODE_DHCP )
    { /* Network port not configured for DHCP */
      if ( eth1_dhcpstate != DHCPSTATE_BOUND )
      { /* Not in a bound state */
        if ( dhcp_bind_network() == L7_SUCCESS )
        {
          if ( eth1_up &&
               eth1_dhcpstate     == DHCPSTATE_BOUND &&
               old_eth1_dhcpstate != eth1_dhcpstate &&
               networkDHCPState   == CONFIG_SET )
          {
		    L7_uchar8 ifname[20];

			sprintf(ifname, "%s0", L7_DTL_PORT_IF);

            /* Set new IP parameters */
            if ( L7_SUCCESS == getParamsFromDhcpStruct((struct bootp *)&eth1_bootp_data, &params) )
            {

			  dhcpConfigSemGet ();
              usmDbAgentBasicConfigNetworkConfigProtocolDesiredGet(USMDB_UNIT_CURRENT, &ibMode2);
              if (ibMode2 == L7_SYSCONFIG_MODE_DHCP) { /* Close race condition */
                /*Set IP Addresss, Default Gateway, Netmask */
                usmDbAgentIpIfAddressSet(USMDB_UNIT_CURRENT,  params.ip);
                usmDbAgentIpIfNetMaskSet(USMDB_UNIT_CURRENT,  params.netMask);
#if defined NETGEAR_GS7xxTR || defined NETGEAR_GS7XXT
                simSetSystemIPGateway(params.gateway, L7_TRUE);
#else
                usmDbAgentIpIfDefaultRouterSet(USMDB_UNIT_CURRENT, params.gateway);
#endif     
              }
			  dhcpConfigSemFree ();

#ifdef _L7_OS_LINUX_
              /*delete route entry for 0.0.0.0 IP address */
              osapimRouteEntryDelete (ifname, 0xffffffff, 0x0, 0xffffffff,
			    0, 0, 1);
#endif
              networkConfigStatus = L7_SYSCONFIG_MODE_COMPLETE;
              networkPortLogStatus = L7_TRUE;
            }
          }
        }

        if ( eth1_dhcpstate == DHCPSTATE_FAILED )
		{
			eth1_dhcpstate= 0;
		}
      }
    }

#if defined NETGEAR_GS7xxTR || defined NETGEAR_GS7XXT || defined NETGEAR_GS1XXT
    osapiSleep(1);
#else
    /* Recheck after some time .....        */
    osapiSleep(10);
#endif
  }

  return L7_SUCCESS;
}


/**************************************************************************
* @purpose  Create configuration protection semaphore.
*
* @comments The DHCP task is started by SIM in phase 2.
*			The configuration protection semaphore allows the task to 
*			read configuration only when the system is in phase 3.
*
* @end
*************************************************************************/
void dhcpConfigSemCreate (void)
{

   dhcpConfigSem = osapiSemaBCreate (OSAPI_SEM_Q_FIFO, OSAPI_SEM_EMPTY);
   if (dhcpConfigSem == L7_NULL) 
   {
	   LOG_ERROR (0);
   }
}

/**************************************************************************
* @purpose  Get the configuration semaphore.
*
* @comments 
*
* @end
*************************************************************************/
void dhcpConfigSemGet (void)
{
  L7_RC_t rc;

   rc = osapiSemaTake (dhcpConfigSem, L7_WAIT_FOREVER);
   if (rc != L7_SUCCESS) 
   {
	   LOG_ERROR (rc);
   }
}

/**************************************************************************
* @purpose  Get the configuration semaphore.
*
* @comments 
*
* @end
*************************************************************************/
void dhcpConfigSemFree (void)
{
	L7_RC_t rc;

	rc = osapiSemaGive (dhcpConfigSem);
	if (rc != L7_SUCCESS) 
	{
		LOG_ERROR (rc);
	}
}

/**************************************************************************
* @purpose  DHCP task startup function.
*
* @comments None.
*
* @end
*************************************************************************/
void dhcpTaskStart()
{

  L7_DHCP_task_id = osapiTaskCreate( "DHCP Client Task", L7_dhcp, 0, 0,
                                     L7_DEFAULT_STACK_SIZE,
                                     L7_DEFAULT_TASK_PRIORITY,
                                     L7_DEFAULT_TASK_SLICE);
  if ( L7_DHCP_task_id == L7_ERROR )
    LOG_MSG("Failed to Create DHCPP Task.\n");
}



/**************************************************************************
* @purpose  Release (and set state to DHCPSTATE_FAILED)for given interface
* @param    Char *intf   Interface Name String
*
* @comments None.
*
* @end
*************************************************************************/
void dhcp_release( void )
{
  if ( eth0_up )
    do_dhcp_release(bspapiServicePortNameGet(),(struct bootp *)&eth0_bootp_data, &eth0_dhcpstate, &eth0_lease);
  if ( eth1_up )
    do_dhcp_release("dtl",(struct bootp *)&eth1_bootp_data, &eth1_dhcpstate, &eth1_lease);

  servPortConfigStatus = L7_SYSCONFIG_MODE_NOT_COMPLETE;
  networkConfigStatus = L7_SYSCONFIG_MODE_NOT_COMPLETE;

}
/* EOF dhcp_support.c*/
