/*
 *  Copyright 2009, Broadcom Corporation
 *   All Rights Reserved.

 *  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Broadcom Corporation;
 *  the contents of this file may not be disclosed to third parties, copied or
 *  duplicated in any form, in whole or in part, without the prior written
 *  permission of Broadcom Corporation.
 */

#include <pkgconf/system.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/typedefs.h>
#include <cyg/hal/bcmendian.h>
#include <cyg/hal/cfe_api.h>
#include <cyg/hal/cfe_error.h>

/*
 * OS needs to store a0~a3 to these variables at its entry point
 *
 * Note: These variables are initialized with non-zero values
 *       since OS could overwrite them if they are in BSS section.
 */
uint32 fw_arg0 = -1;
uint32 fw_arg1 = -1;
uint32 fw_arg2 = -1;
uint32 fw_arg3 = -1;

void
prom_init(void)
{
	uint32_t cfe_ept, cfe_handle;
	unsigned int cfe_eptseal;
	int argc = fw_arg0;
	char **envp = (char **) fw_arg2;
	int *prom_vec = (int *) fw_arg3;
	
	/*
	 * Check if a loader was used; if NOT, the 4 arguments are
	 * what CFE gives us (handle, 0, EPT and EPTSEAL)
	 */
	if (argc < 0) {
		cfe_handle = (uint32_t)(long)argc;
		cfe_ept = (long)envp;
		cfe_eptseal = (uint32_t)(unsigned long)prom_vec;
	} else {
		if ((int32_t)(long)prom_vec < 0) {
			/*
			 * Old loader; all it gives us is the handle,
			 * so use the "known" entrypoint and assume
			 * the seal.
			 */
			cfe_handle = (uint32_t)(long)prom_vec;
			cfe_ept = (uint32_t)((int32_t)0x9fc00500);
			cfe_eptseal = CFE_EPTSEAL;
		} else {
			/*
			 * Newer loaders bundle the handle/ept/eptseal
			 * Note: prom_vec is in the loader's useg
			 * which is still alive in the TLB.
			 */
			cfe_handle = (uint32_t)((int32_t *)prom_vec)[0];
			cfe_ept = (uint32_t)((int32_t *)prom_vec)[2];
			cfe_eptseal = (unsigned int)((uint32_t *)prom_vec)[3];
		}
	}
	if (cfe_eptseal != CFE_EPTSEAL) {
		/* too early for panic to do any good */
		return;
	}
	cfe_init(cfe_handle, cfe_ept);
}
