//==========================================================================
//
//      lib/tftp_server.c
//
//      TFTP server support
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    gthomas
// Contributors: gthomas, hmt, andrew.lunn@ascom.ch (Andrew Lunn)
// Date:         2000-04-06
// Purpose:      
// Description:  
//              
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <arpa/tftp.h>        // TFTP protocol definitions
#ifndef _L7_OS_ECOS_
#include <tftp_support.h>     // TFTPD structures
#else
#include <network.h>          // Basic networking support
#include <cyg/kernel/kapi.h>
#include "include/tftp_support.h"
#endif
#include <stdlib.h>           // For malloc
#include<stdio.h>
#include<string.h>
#include "l7_common.h"
#include "l7_common_l3.h"
#include "l7_socket.h"
#include "osapi_sockdefs.h"
#include "osapi_support.h"

#define nCYGOPT_NET_TFTP_SERVER_INSTRUMENT

#undef TFTP_DEBUG

struct subinfo {
    int rx;
    int rx_repeat;
    int rx_skip;
    int send;
    int resend;
};

struct info {
    struct subinfo ack, data;
    int err_send;
    int total_transactions;
};


static struct info tftp_server_instrument = {
    { 0,0,0,0,0 },
    { 0,0,0,0,0 },
    0, 0,
};

static char *TFTP_tag = "TFTPD";
static L7_int32 tftpserver_taskId = L7_ERROR;
struct tftp_server {
    char                 *tag;
    int                  port;
    struct tftpd_fileops *ops;
};

static char * errmsg[] = {
  "Undefined error code",               // 0 nothing defined
  "File not found",                     // 1 TFTP_ENOTFOUND 
  "Access violation",                   // 2 TFTP_EACCESS   
  "Disk full or allocation exceeded",   // 3 TFTP_ENOSPACE  
  "Illegal TFTP operation",             // 4 TFTP_EBADOP    
  "Unknown transfer ID",                // 5 TFTP_EBADID    
  "File already exists",                // 6 TFTP_EEXISTS   
  "No such user",                       // 7 TFTP_ENOUSER   
};

/* add status for check */
static tftp_status_t tftp_status = TFTP_STATUS_NONE;

/* Send an error packet to the client */
static void tftpd_send_error(int s, struct tftphdr * reply, int err,
	                     L7_sockaddr_in_t *from_addr, int from_len)
{
    L7_uint32 len;
    
    reply->th_opcode = htons(ERROR);
    reply->th_code = htons(err);
    if ( (0 > err) || (sizeof(errmsg)/sizeof(errmsg[0]) <= err) )
        err = 0; // Do not copy a random string from hyperspace
    strcpy(reply->th_msg, errmsg[err]);
    osapiSocketSendto(s, (L7_uchar8 *)reply, 4+strlen(reply->th_msg)+1, 0, 
	   (L7_sockaddr_t *)from_addr, from_len, &len);
}

//
// Receive a file from the client
//
static void
tftpd_write_file(struct tftp_server *server,
                 struct tftphdr *hdr, 
                 L7_sockaddr_in_t *from_addr, int from_len)
{
  char data_out[SEGSIZE+sizeof(struct tftphdr)];
  char data_in[SEGSIZE+sizeof(struct tftphdr)];
  struct tftphdr *reply = (struct tftphdr *)data_out;
  struct tftphdr *response = (struct tftphdr *)data_in;
  int fd, len, ok, tries, closed, data_len, s,out_len;
  unsigned short block;
  struct timeval timeout;
  fd_set fds;
  int total_timeouts = 0;
  L7_sockaddr_in_t client_addr, local_addr;
  int client_len;
#ifdef TFTP_DEBUG
  FILE*          fp;
  int            tempFd;
  L7_timespec    currentTime;
  L7_uchar8      debug[64];
#endif

  if (osapiSocketCreate(L7_AF_INET, L7_SOCK_DGRAM, 0,&s) == L7_FAILURE)
  {
    sysapiPrintf("TFTPD: can't open socket for 'write_file'\n");
    return;
  }
    
  memset((char *)&local_addr, 0, sizeof(local_addr));
  local_addr.sin_family = L7_AF_INET;
  local_addr.sin_len = sizeof(local_addr);
  local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  
  /* fit for netgear discovery smartwizard 
     local_addr.sin_port = htons(INADDR_ANY); 
   */
  local_addr.sin_port = htons(69);

  if (osapiSocketBind(s, (L7_sockaddr_t *)&local_addr, sizeof(local_addr)) != L7_SUCCESS)
  {
    // Problem setting up my end
    sysapiPrintf("TFTPD: can't bind to reply port for 'write_file'\n");
    osapiSocketClose(s);
    return;
  }

#if !(defined NETGEAR_GS7XXT || defined NETGEAR_GS1XXT)
  if ((fd = (server->ops->open)(hdr->th_stuff, O_WRONLY)) < 0) 
#else
  fd = (server->ops->open)(hdr->th_stuff, O_WRONLY);
  if (fd == L7_ERROR || fd == L7_NULL)
#endif
  {
    tftpd_send_error(s,reply,TFTP_ENOTFOUND,from_addr, from_len); 
    tftp_status = TFTP_STATUS_FAIL;
    osapiSocketClose(s);
    return;
  }

  ok = L7_TRUE;
  closed = L7_FALSE;
  block = 0;
    
#ifdef TFTP_DEBUG
  if (osapiFsCreateFile("/mnt/download/debug-log") != L7_SUCCESS)
    sysapiPrintf("Debug log File creation failed\n");

  tempFd = osapiFsOpen("/mnt/download/debug-log");
  if (tempFd == L7_ERROR)
    sysapiPrintf("Error opeing debug file");

  fp=fdopen(tempFd, "a+");
  if (fp==L7_NULLPTR)
  { 
    sysapiPrintf("Error while opening Debug file");
  }
#endif

  while (ok) 
  {
    // Send ACK telling client he can send data
    reply->th_opcode = htons(ACK);
    reply->th_block = htons(block++); // postincrement
    
    for (tries = 0;  tries < TFTP_RETRIES_MAX;  tries++) 
    {
      tftp_server_instrument.ack.send++;
            
      repeat_select:

#ifdef TFTP_DEBUG
      osapiUpTime(&currentTime); 
      memset(debug, 0x00, sizeof(debug));
      snprintf(debug, sizeof(debug), "%d:%d Sending ACK for block %d\n",currentTime.minutes,currentTime.seconds,block);
      fileWrite(tempFd,debug,sizeof(debug),"ASCII",fp);
#endif
      if (osapiSocketSendto(s, (L7_uchar8 *)reply, 4, 0,(L7_sockaddr_t *)from_addr, from_len,&out_len) != L7_SUCCESS ) 
      {      
        sysapiPrintf("%s : %d osapiSocketSendto error in tftpd_write_file()\n", __FUNCTION__,__LINE__);
      }
      timeout.tv_sec = TFTP_TIMEOUT_PERIOD;
      timeout.tv_usec = 0;
      FD_ZERO(&fds);
      FD_SET(s, &fds);
            
      if (select(s+1, &fds, 0, 0, &timeout) <= 0) 
      {
        if (++total_timeouts > TFTP_TIMEOUT_MAX) 
        {
          tftp_server_instrument.err_send++;
          tftpd_send_error(s,reply,TFTP_EBADOP,from_addr, from_len);
          ok = L7_FALSE;
          tftp_status = TFTP_STATUS_FAIL;
          break;
        }
        tftp_server_instrument.ack.resend++;
        sysapiPrintf("select: failed retry count %d\n",tries);
        continue; // retry the send, using up one retry.
      }
      
      // Some data has arrived
      data_len = sizeof(data_in);
      client_len = sizeof(client_addr);
            
      if (osapiSocketRecvfrom(s, data_in, data_len,0,(L7_sockaddr_t *)&client_addr, &client_len,&data_len) != L7_SUCCESS) 
      {
        // What happened?  No data here!
        tftp_server_instrument.ack.resend++;
        sysapiPrintf("osapiSocketRecvfrom failed\n");
        continue; // retry the send, using up one retry.
      }

#ifdef TFTP_DEBUG
      osapiUpTime(&currentTime); 
      memset(debug, 0x00, sizeof(debug));
      snprintf(debug, sizeof(debug), "%d:%d Received TFTP block %d\n",
               currentTime.minutes, currentTime.seconds, ntohs(response->th_block));
      fileWrite(tempFd,debug,sizeof(debug),"ASCII",fp);
#endif

      if (ntohs(response->th_opcode) == DATA && ntohs(response->th_block) < block) 
      {
        tftp_server_instrument.data.rx_repeat++;
        // Then it is repeat DATA with an old block; listen again,
        // but do not repeat sending the current ack, and do not
        // use up a retry count.  (we do re-send the ack if
        // subsequently we time out)
        goto repeat_select;
      }

      if (ntohs(response->th_opcode) == DATA && ntohs(response->th_block) == block) 
      {
        tftp_server_instrument.data.rx++;
        // Good data - write to file

#ifdef TFTP_DEBUG
       memset(debug, 0x00, sizeof(debug));
       snprintf(debug, sizeof(debug), "Writing %d bytes to flash\n",data_len-4);
       fileWrite(tempFd,debug,sizeof(debug),"ASCII",fp);
#endif
       len = (server->ops->write)(fd, response->th_data, data_len-4);

#ifdef TFTP_DEBUG
       memset(debug, 0x00, sizeof(debug));
       snprintf(debug, sizeof(debug), "%d bytes written successfully\n",len);
       fileWrite(tempFd,debug,sizeof(debug),"ASCII",fp);
#endif
       if(len == -2) 
       {
         // illegal TFTP operation
         tftpd_send_error(s,reply,TFTP_EBADOP, from_addr, from_len);     
         ok = L7_FALSE;  // Give up
         tftp_status = TFTP_STATUS_FAIL;
         break; // out of the retries loop
       }
       
       if (len < 0)
       {
         sysapiPrintf("error while writing the file return code %d",len);
         tftpd_send_error(s,reply,TFTP_ENOSPACE, from_addr, from_len);
         ok = L7_FALSE;  // Give up
         tftp_status = TFTP_STATUS_FAIL;
         break; // out of the retries loop
       }

       if (data_len < (SEGSIZE+4)) 
       {
         // End of file
         closed = L7_TRUE;
         ok = L7_FALSE;
                
#if !(defined NETGEAR_GS7XXT || defined NETGEAR_GS1XXT)
         if ((server->ops->close)(fd) == -1) 
#else
         if ((server->ops->close)(fd, L7_TRANSFER_DIRECTION_DOWN) == -1) 
#endif
         {
           tftpd_send_error(s,reply,TFTP_EACCESS, from_addr, from_len);
           tftp_status = TFTP_STATUS_FAIL;
           break;  // out of the retries loop
         }
                    
         // Exception to the loop structure: we must ACK the last
         // packet, the one that implied EOF:
         tftp_server_instrument.ack.send++;
         reply->th_opcode = htons(ACK);
         reply->th_block = htons(block++); // postincrement
         osapiSocketSendto(s, (L7_uchar8 *)reply, 4, 0, (L7_sockaddr_t *)from_addr, from_len,&len);
         tftp_status = TFTP_STATUS_SUCCESS;
         break; // out of the retries loop
       }
       // Happy!  Break out of the retries loop.
       break;
     }
    
      // Otherwise, we got something we do not understand!  So repeat
      // sending the current ACK, and use up a retry count.
      if ((ntohs(response->th_opcode) == DATA) )
       tftp_server_instrument.data.rx_skip++;
    
      tftp_server_instrument.ack.resend++;
    } // End of the retries loop.

    if (TFTP_RETRIES_MAX <= tries) 
    {
      tftp_server_instrument.err_send++;
      tftpd_send_error(s,reply,TFTP_EBADOP,from_addr, from_len);
      ok = L7_FALSE;
      tftp_status = TFTP_STATUS_FAIL;
    }
  }
  osapiSocketClose(s);
  if (!closed) 
  {
#if !(defined NETGEAR_GS7XXT || defined NETGEAR_GS1XXT)
    (server->ops->close)(fd);
#else
    (server->ops->close)(fd, L7_TRANSFER_DIRECTION_DOWN);
#endif
  }
}

//
// Send a file to the client
//
static void
tftpd_read_file(struct tftp_server *server,
                struct tftphdr *hdr, 
                L7_sockaddr_in_t *from_addr, int from_len)
{
    char data_out[SEGSIZE+sizeof(struct tftphdr)];
    char data_in[SEGSIZE+sizeof(struct tftphdr)];
    struct tftphdr *reply = (struct tftphdr *)data_out;
    struct tftphdr *response = (struct tftphdr *)data_in;
    int fd, len, tries, ok, data_len, s,out_len;
    unsigned short block;
    struct timeval timeout;
    fd_set fds;
    int total_timeouts = 0;
    L7_sockaddr_in_t client_addr, local_addr;
    int client_len;

    if (osapiSocketCreate(L7_AF_INET, L7_SOCK_DGRAM, 0,&s) == L7_FAILURE)
    {
        sysapiPrintf("TFTPD: can't open socket for 'read_file'\n");
        return;
    }
    memset((char *)&local_addr, 0, sizeof(local_addr));
    local_addr.sin_family = L7_AF_INET;
    local_addr.sin_len = sizeof(local_addr);
    local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    local_addr.sin_port = htons(INADDR_ANY);

    if (osapiSocketBind(s, (L7_sockaddr_t *)&local_addr, sizeof(local_addr)) != L7_SUCCESS)
    {
        // Problem setting up my end
        sysapiPrintf("TFTPD: can't bind to reply port for 'read_file'\n");
        osapiSocketClose(s);
        return;
    }
    if ((fd = (server->ops->open)(hdr->th_stuff, O_RDONLY)) < 0) {
        tftpd_send_error(s,reply,TFTP_ENOTFOUND,from_addr, from_len);
        osapiSocketClose(s);
        return;
    }
    block = 0;
    ok = L7_TRUE;
    while (ok) {
        // Read next chunk of file
        len = (server->ops->read)(fd, reply->th_data, SEGSIZE);
        reply->th_block = htons(++block); // preincrement
        reply->th_opcode = htons(DATA);
        for (tries = 0;  tries < TFTP_RETRIES_MAX;  tries++) {
            tftp_server_instrument.data.send++;
            
            if (osapiSocketSendto(s, (L7_uchar8 *)reply, 4+len, 0,
                                  (L7_sockaddr_t *)from_addr, from_len, &out_len) != L7_SUCCESS) 
            {
                // Something went wrong with the network!
                ok = L7_FALSE;
                tftp_status = TFTP_STATUS_FAIL;
                break;
            }
        repeat_select:
            timeout.tv_sec = TFTP_TIMEOUT_PERIOD;
            timeout.tv_usec = 0;
            FD_ZERO(&fds);
            FD_SET(s, &fds);
            if (select(s+1, &fds, 0, 0, &timeout) <= 0) {
                if (++total_timeouts > TFTP_TIMEOUT_MAX) {
                    tftp_server_instrument.err_send++;
 		    tftpd_send_error(s,reply,TFTP_EBADOP,from_addr, from_len);
                    ok = L7_FALSE;
                    tftp_status = TFTP_STATUS_FAIL;
                    break;
                }
                tftp_server_instrument.data.resend++;
                continue; // retry the send, using up one retry.
            }
            data_len = sizeof(data_in);
            client_len = sizeof(client_addr);
          if (osapiSocketRecvfrom(s, data_in, data_len,0,(L7_sockaddr_t *)&client_addr, &client_len,&data_len) != L7_SUCCESS)
          {
                // What happened?  Maybe someone lied to us...
                tftp_server_instrument.data.resend++;
                continue; // retry the send, using up one retry.
            }
            if ((ntohs(response->th_opcode) == ACK) &&
                (ntohs(response->th_block) < block)) {
                tftp_server_instrument.ack.rx_repeat++;
                // Then it is a repeat ACK for an old block; listen again,
                // but do not repeat sending the current block, and do not
                // use up a retry count.  (we do re-send the data if
                // subsequently we time out)
                goto repeat_select;
            }
            if ((ntohs(response->th_opcode) == ACK) &&
                (ntohs(response->th_block) == block)) {
                tftp_server_instrument.ack.rx++;
                // Happy!  Break out of the retries loop.
                break;
            }
            // Otherwise, we got something we do not understand!  So repeat
            // sending the current block, and use up a retry count.
            if ( (ntohs(response->th_opcode) == ACK) )
                tftp_server_instrument.ack.rx_skip++;
            tftp_server_instrument.data.resend++;
        } // End of the retries loop.
        if (TFTP_RETRIES_MAX <= tries) {
            tftp_server_instrument.err_send++;
            tftpd_send_error(s,reply,TFTP_EBADOP,from_addr, from_len);
            ok = L7_FALSE;
            tftp_status = TFTP_STATUS_FAIL;
        }
        if (len < SEGSIZE) {
            tftp_status = TFTP_STATUS_SUCCESS;
            break; // That's end of file then.
        }
    }
    osapiSocketClose(s);
#if !(defined NETGEAR_GS7XXT || defined NETGEAR_GS1XXT)
    (server->ops->close)(fd);
#else
    (server->ops->close)(fd, L7_TRANSFER_DIRECTION_UP);
#endif
}

//
// Actual TFTP server
//

static void
tftpd_server(L7_uint32 numArgs, struct tftp_server *p)
{
    struct tftp_server *server;
    int s;
    int data_len, recv_len, from_len;
    L7_sockaddr_in_t local_addr, from_addr;
    struct servent *server_info;
    char data[SEGSIZE+sizeof(struct tftphdr)];
    struct tftphdr *hdr = (struct tftphdr *)data;

    server = p; 
  
    // Otherwise routine printfs fail the test - interrupts disabled too long.
#ifdef TFTP_DEBUG
    sysapiPrintf("TFTPD [%x]: port %d\n", p, server->port);
#endif
   
    // Set up port
    if (server->port == 0) 
    {
      server_info = getservbyname("tftp", "udp");
    
      if (server_info == (struct servent *)0) 
      {
        sysapiPrintf("TFTPD: can't get TFTP service information\n");
        return;
      }
      // Network order in info; host order in server:
      server->port = ntohs( server_info->s_port );
    }

    while (L7_TRUE) 
    {
       struct info o = tftp_server_instrument;
        // Create socket
       if(osapiSocketCreate(L7_AF_INET, L7_SOCK_DGRAM, 0,&s) == L7_FAILURE)
       {
         sysapiPrintf("TFTPD [%x]: can't open socket\n", p);
         return;
       }
 
       memset((char *)&local_addr, 0, sizeof(local_addr));
       local_addr.sin_family = L7_AF_INET;
       local_addr.sin_len = sizeof(local_addr);
       local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
       local_addr.sin_port = htons(server->port);

       if (osapiSocketBind(s, (L7_sockaddr_t *)&local_addr, sizeof(local_addr)) != L7_SUCCESS) 
       {
         // Problem setting up my end
         osapiSocketClose(s);
#ifdef CYGSEM_TFTP_SERVER_MULTITHREADED
#ifndef CYGPKG_NET_TESTS_USE_RT_TEST_HARNESS
         sysapiPrintf("TFTPD [%x]: waiting to bind to service port\n", p);
#endif
         // Wait until the socket is free...
         cyg_semaphore_wait( &tftp_server_sem );
         continue; // try re-opening and rebinding the socket.
#else
         sysapiPrintf("TFTPD [%x]: can't bind to service port\n", p);
         return;
#endif
       }
       recv_len = sizeof(data);
       from_len = sizeof(from_addr);

       if (osapiSocketRecvfrom(s, (L7_uchar8 *)hdr, recv_len, 0,(L7_sockaddr_t *)&from_addr, &from_len, &data_len) != L7_SUCCESS)
       {
         /* receive error */
         sysapiPrintf("osapiSocketRecvfrom error in tftpd_server()");
       }
      
        osapiSocketClose(s); // so that other servers can bind to the TFTP socket
#ifdef CYGSEM_TFTP_SERVER_MULTITHREADED
        // The socket is free...
        cyg_semaphore_post( &tftp_server_sem );
#endif

        if ( data_len < 0) 
        {
            sysapiPrintf("TFTPD [%x]: can't read request\n", p);
        } 
        else 
        {
#ifdef TFTP_DEBUG 
            sysapiPrintf("TFTPD [%x]: received %x from %s:%d\n", p,
                        ntohs(hdr->th_opcode), osapiInet_ntoa(from_addr.sin_addr.s_addr),
                        ntohs(from_addr.sin_port));
#endif
          switch (ntohs(hdr->th_opcode)) 
          {
            case WRQ:
                tftpd_write_file(server, hdr, &from_addr, from_len);
                break; 
            case RRQ:
                tftpd_read_file(server, hdr, &from_addr, from_len);
                break;
            case ACK:
            case DATA:
            case ERROR:
                // Ignore
                break;
            default:
                sysapiPrintf("TFTPD [%x]: bogus request %x from %s:%d\n", p,
                            ntohs(hdr->th_opcode),
                            osapiInet_ntoa(from_addr.sin_addr.s_addr),
                            from_addr.sin_port );
                tftpd_send_error(s,hdr,TFTP_EBADOP,&from_addr,from_len);
            }
        }
        tftp_server_instrument.total_transactions++;

        o.data.rx        -= tftp_server_instrument.data.rx       ;
        o.data.rx_repeat -= tftp_server_instrument.data.rx_repeat;
        o.data.rx_skip   -= tftp_server_instrument.data.rx_skip  ;
        o.data.send      -= tftp_server_instrument.data.send     ;
        o.data.resend    -= tftp_server_instrument.data.resend   ;

        o.ack.rx         -= tftp_server_instrument.ack.rx        ;
        o.ack.rx_repeat  -= tftp_server_instrument.ack.rx_repeat ;
        o.ack.rx_skip    -= tftp_server_instrument.ack.rx_skip   ;
        o.ack.send       -= tftp_server_instrument.ack.send      ;
        o.ack.resend     -= tftp_server_instrument.ack.resend    ;

        o.err_send       -= tftp_server_instrument.err_send      ;

        if ( o.data.rx        ) sysapiPrintf( "data rx       %4d\n", -o.data.rx        );
        if ( o.data.rx_repeat ) sysapiPrintf( "data rx_repeat%4d\n", -o.data.rx_repeat );
        if ( o.data.rx_skip   ) sysapiPrintf( "data rx_skip  %4d\n", -o.data.rx_skip   );
        if ( o.data.send      ) sysapiPrintf( "data send     %4d\n", -o.data.send      );
        if ( o.data.resend    ) sysapiPrintf( "data resend   %4d\n", -o.data.resend    );

        if ( o.ack.rx        ) sysapiPrintf( " ack rx       %4d\n", -o.ack.rx        );
        if ( o.ack.rx_repeat ) sysapiPrintf( " ack rx_repeat%4d\n", -o.ack.rx_repeat );
        if ( o.ack.rx_skip   ) sysapiPrintf( " ack rx_skip  %4d\n", -o.ack.rx_skip   );
        if ( o.ack.send      ) sysapiPrintf( " ack send     %4d\n", -o.ack.send      );
        if ( o.ack.resend    ) sysapiPrintf( " ack resend   %4d\n", -o.ack.resend    );

        if ( o.err_send      ) sysapiPrintf( "*error sends  %4d\n", -o.err_send      );
    /* 
     *  For older version of nsdp(discovery) it have't sence becouse after image
     *  ubgrade had been finished - reboot is the next mandatory action.
     *
     *  For never one (v4) this is bug because disallow multiple tftp transactions.
     */
#if 0
    return;
#endif /*0*/
    }
}

//
// This function is used to create a new server [thread] which supports
// the TFTP protocol on the given port.  A server 'id' will be returned
// which can later be used to destroy the server.  
//
// Note: all [memory] resources for the server thread will be allocated
// dynamically.  If there are insufficient resources, an error will be
// returned.
//
int 
l7_tftpd_start(int port, struct tftpd_fileops *ops)
{
    struct tftp_server *server;

    if ((server = malloc(sizeof(struct tftp_server)))) 
    {
      server->tag = TFTP_tag;
      server->port = port;
      server->ops = ops;
     
      if((tftpserver_taskId = osapiTaskCreate("TFTP server", tftpd_server, 1, server, L7_DEFAULT_STACK_SIZE,
                     L7_DEFAULT_TASK_PRIORITY, L7_DEFAULT_TASK_SLICE)) == L7_ERROR)
      {
         sysapiPrintf("tftpd_server : task Create Failed\n");
         LOG_MSG("Task create failed\n");
      }
    }

    if(server != 0) 
    {
        tftp_status = TFTP_STATUS_PROCESS;
    }
        
    return (int)server;
}

//
// Destroy a TFTP server, using a previously created server 'id'.
//
int 
l7_tftpd_stop(int p)
{
    struct tftp_server *server = (struct tftp_server *)p;

    if(server == L7_NULLPTR) 
    {
      return 0;
    }

    // Simple sanity check
    if (server->tag == TFTP_tag) 
    {
       if (tftpserver_taskId != L7_ERROR)
       {
#if 0 /* This call fail and restart switch */
         osapiTaskDelete(tftpserver_taskId); 
         tftpserver_taskId = L7_ERROR;
#endif 
       }
    }

    tftp_status = TFTP_STATUS_NONE;
    return 0;
}

int
l7_tftpd_status(void) {
    return tftp_status;
}

/*********************************************************************
* @purpose  Sets TFTP server status
*
* @param    status  @b{(in)} tftp server status
*
* @returns  L7_SUCCESS
*
* @comments
*
* @end
*********************************************************************/
void tftpdStatusSet(tftp_status_t status)
{
  tftp_status = status;
}
// EOF tftp_server.c
